"""Tests for wrapper helpers for making DataStreams play nicely with PyTorch DataLoaders.
"""
# Standard
from unittest import mock

# Third Party
from torch.utils.data._utils import worker
import pytest

# First Party
from caikit.core.data_model import DataStream

# Local
from caikit_nlp.toolkit.data_stream_wrapper import SimpleIterableStreamWrapper
from tests.fixtures import requires_determinism

# Sample data to load via PyTorch
SAMPLE_DATA = [{"label": "foo"}, {"label": "foo"}, {"label": "bar"}, {"label": "bar"}]
SAMPLE_STREAM = DataStream.from_iterable(SAMPLE_DATA)
NUM_CYCLES = 10


def test_without_shuffling():
    """Ensure that we can build a datastream & load it in a data loader without shuffling."""
    test_results = []
    # Get the IDs of all objects in the stream
    get_stream_id_order = lambda s: [id(datum) for datum in s]
    # Compare the data stream at two different iteration points; here, we
    # produce True if two streams have the same objects in the same order
    have_same_id_order = lambda id_set1, id_set2: all(
        [datum1 == datum2 for datum1, datum2 in zip(id_set1, id_set2)]
    ) and len(id_set1) == len(id_set2)

    # NOTE - a buffer size of 1 is a noop; the shuffle operation just gets the current element
    wrapper = SimpleIterableStreamWrapper(stream=SAMPLE_STREAM, shuffle=False)
    # Cycle through NUM_CYCLES times & ensure that the order of our objects does not change
    initialize_order = get_stream_id_order(wrapper)
    for _ in range(NUM_CYCLES):
        cycle_ids = get_stream_id_order(wrapper)
        test_res = have_same_id_order(initialize_order, cycle_ids)
        test_results.append(test_res)
    assert all(test_results)


def test_shuffle_full_buffer(requires_determinism):
    """Ensure that we can build a datastream & shuffle it all in memory on each iteration."""
    test_results = []
    # Get the IDs of all objects in the stream
    get_stream_id_order = lambda s: [id(datum) for datum in s]
    # Compare the data stream at two different iteration points; here, we
    # produce True if two streams have the same objects in the same order
    have_same_id_order = lambda id_set1, id_set2: all(
        [datum1 == datum2 for datum1, datum2 in zip(id_set1, id_set2)]
    ) and len(id_set1) == len(id_set2)

    # NOTE - a buffer size of 1 is a noop; the shuffle operation just gets the current element
    wrapper = SimpleIterableStreamWrapper(
        stream=SAMPLE_STREAM, shuffle=True, buffer_size=len(SAMPLE_STREAM)
    )
    # Cycle through NUM_CYCLES times & ensure that the order of our objects DOES change sometimes
    initialize_order = get_stream_id_order(wrapper)
    for _ in range(NUM_CYCLES):
        cycle_ids = get_stream_id_order(wrapper)
        test_res = have_same_id_order(initialize_order, cycle_ids)
        test_results.append(test_res)
    assert not all(test_results)


# def test_iter_with_multi_worker(requires_determinism):
#     """Ensure that we are able to iterate properly over data in case of workers
#     managed by torch"""

#     test_results = []
#     # Get the IDs of all objects in the stream
#     get_stream_id_order = lambda s: [id(datum) for datum in s]
#     # Compare the data stream at two different iteration points; here, we
#     # produce True if two streams have the same objects in the same order
#     have_same_id_order = lambda id_set1, id_set2: all(
#         [datum1 == datum2 for datum1, datum2 in zip(id_set1, id_set2)]
#     ) and len(id_set1) == len(id_set2)

#     dummy_worker_info = worker.WorkerInfo(
#         id=1,
#         num_workers=2,
#         seed=7,
#     )

#     with mock.patch.object(worker, '_worker_info', dummy_worker_info):
#         wrapper = SimpleIterableStreamWrapper(stream=SAMPLE_STREAM, shuffle=False)
#         initialize_order = get_stream_id_order(wrapper)
#         for _ in range(NUM_CYCLES):
#             cycle_ids = get_stream_id_order(wrapper)
#             test_res = have_same_id_order(initialize_order, cycle_ids)
#             test_results.append(test_res)
#         assert not all(test_results)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains sequence classification, compatible with transformer
SequenceClassification modules. At this time this module is only designed
for inference"""
# Standard
from typing import Dict, List, Union

# Third Party
import torch

# First Party
from caikit.core.exceptions import error_handler
from caikit.core.modules import ModuleBase, ModuleLoader, ModuleSaver, module
from caikit.interfaces.nlp.data_model import ClassificationResult, ClassificationResults
from caikit.interfaces.nlp.tasks import TextClassificationTask
import alog

# Local
from ...resources.pretrained_model.hf_auto_seq_classifier import (
    HFAutoSequenceClassifier,
)

log = alog.use_channel("SEQ_CLASS")
error = error_handler.get(log)


@module(
    id="d21107ca-d579-4321-aedd-3099a526e0dd",
    name="Sequence classification",
    version="0.1.0",
    task=TextClassificationTask,
)
class SequenceClassification(ModuleBase):

    ################################ Constructor #################################################

    def __init__(
        self,
        resource: HFAutoSequenceClassifier,
    ):
        super().__init__()
        error.type_check(
            "<NLP74125820E>",
            HFAutoSequenceClassifier,
            resource=resource,
        )
        self.resource = resource
        self.tokenizer = resource.tokenizer
        self.model = resource.model

    ################################## API functions #############################################

    def run(self, text: str) -> ClassificationResults:
        """Run the sequence classification.
            NOTE: This will truncate sequences that are too long for model

        Args:
            text: str
                Input string to be classified

        Returns:
            ClassificationResults
        """
        scores_dict = self._get_scores(text)
        # Re-organize scores_dict - for one text, this is just the first score
        return SequenceClassification._process_predictions(scores_dict, text_idx=0)

    def run_batch(self, texts: List[str]) -> List[ClassificationResults]:
        """Run the sequence classification on batch, truncates sequences too long for model

        Args:
            text: List[str]
                Input strings to be classified

        Returns:
            List[ClassificationResults]
        """
        scores_dict = self._get_scores(texts)
        num_texts = len(texts)

        # Re-organize scores_dict for each example
        # We could eventually consider whether or not to sort classifications by scores
        # but avoiding this prescription here for now
        classification_predictions = []
        for text_idx in range(num_texts):
            classification_prediction = SequenceClassification._process_predictions(
                scores_dict, text_idx
            )
            classification_predictions.append(classification_prediction)
        return classification_predictions

    def save(self, model_path: str):
        """Save model in target path

        Args:
            model_path: str
                Path to store model artifact(s)
        """
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )

        with module_saver:
            module_saver.save_module(self.resource, "sequence_classifier")

    @classmethod
    def load(cls, model_path: str) -> "SequenceClassification":
        """Load a sequence classification model

        Args:
            model_path: str
                Path to the model to be loaded.

        Returns:
            SequenceClassification
                Instance of this class built from the on disk model.
        """
        loader = ModuleLoader(model_path)
        resource = loader.load_module("sequence_classifier")
        return cls(resource=resource)

    @classmethod
    def bootstrap(cls, base_model_path: str) -> "SequenceClassification":
        """Bootstrap a HuggingFace transformer-based sequence classification model

        Args:
            base_model_path: str
                Path to the model to be loaded.
        """
        # Note: Must provide path to tokenizer if model_name is a path
        # for resource use
        resource = HFAutoSequenceClassifier.bootstrap(
            model_name=base_model_path, tokenizer_name=base_model_path
        )
        return cls(
            resource=resource,
        )

    ################################## Private Functions #########################################

    def _get_scores(self, text: Union[str, List[str]]):
        """Run tokenizer and model to get scores on text(s)

        Args:
            text: Union[str, List[str]]
                Input string(s) to be used

        Returns:
            scores_dict
                Dict with key label, and values as the array of scores,
                each corresponding to text(s)
        """
        # NOTE: no explicit GPU support at this time

        # Apply tokenizer
        tokenized_text = self.tokenizer(
            text,
            padding="max_length",
            truncation=True,
            return_tensors="pt",  # PyTorch
        )
        # NOTE: no truncation warning given at this time
        # difficult to detect if truncation happened since padding also occurs
        with torch.no_grad():
            logits = self.model(**tokenized_text).logits

        if not self.model.config.id2label:
            log.warning(
                "<NLP31047577W>",
                "No id2label provided in model config. Defaulting to numeric labels",
            )

        softmax = torch.nn.Softmax(dim=1)
        raw_scores = softmax(logits)
        scores = raw_scores.numpy()
        num_labels = self.model.num_labels
        num_texts = 1  # str
        if isinstance(text, List):
            num_texts = len(text)
        error.value_check(
            "<NLP33929938E>",
            scores.shape == (num_texts, num_labels),
            "model logits expected to be of shape (num_texts, num_labels)",
        )

        scores_dict = {}
        for label_idx in range(num_labels):
            if self.model.config.id2label:
                label = self.model.config.id2label[label_idx]
            else:
                label = label_idx
            label_scores = scores[:, label_idx]
            scores_dict[label] = label_scores
        return scores_dict

    @staticmethod
    def _process_predictions(scores_dict: Dict, text_idx: int) -> ClassificationResults:
        """Process dictionary of label: scores to ClassificationResults

        Args:
            scores_dict: Dict
                Dict with key label, and values as the array of scores,
                each corresponding to text(s)
            text_idx: int
                Integer index of text in batch

        Returns:
            ClassificationResults
        """
        error.type_check("<NLP40517898E>", Dict, scores_dict=scores_dict)
        classification_list = []
        for label, score_array in scores_dict.items():
            # NOTE: labels are expected to be str, especially for config
            classification_list.append(
                ClassificationResult(label=str(label), score=score_array[text_idx])
            )
        return ClassificationResults(results=classification_list)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import Optional, Union

# Third Party
import torch

# First Party
from caikit import get_config
from caikit.core.exceptions import error_handler
import alog

log = alog.use_channel("DATA_UTIL")
error = error_handler.get(log)


def str_to_torch_dtype(dtype_str: str) -> torch.dtype:
    """Given a string representation of a Torch data type, convert it to the actual torch dtype.

    Args:
        dtype_str: String representation of Torch dtype to be used; this should be an attr
        of the torch library whose value is a dtype.

    Returns:
        torch.dtype
            Data type of the Torch class being used.
    """
    dt = getattr(torch, dtype_str, None)
    if not isinstance(dt, torch.dtype):
        error("<NLP85554812E>", ValueError(f"Unrecognized data type: {dtype_str}"))
    return dt


def get_torch_dtype(dtype: Optional[Union[str, torch.dtype]] = None) -> torch.dtype:
    """Get the Torch data type to be used for interacting with a model.

    Args:
        dtype: Optional[Union[str, torch.dtype]]
            If dtype is a torch.dtype, returns it; if it's a string, grab it from the Torch lib.
            If None is provided, fall back to the default type in config, which can be
            overridden via environment variable.

    Returns:
        torch.dtype
            Torch data type to be used.
    """
    error.type_check("<NLP84274822E>", torch.dtype, str, dtype=dtype, allow_none=True)
    # If a Torch dtype is passed, nothing to do
    if isinstance(dtype, torch.dtype):
        return dtype
    # If None/empty str was provided, fall back to config / env var override
    if not dtype:
        return str_to_torch_dtype(get_config().torch_dtype)
    # Otherwise convert it from a string
    return str_to_torch_dtype(dtype)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Standard
from functools import cached_property
from typing import Iterable, List, Optional, Tuple, Union
import os

# Third Party
import numpy as np

# First Party
from caikit.core.exceptions import error_handler
from caikit.core.module_backends import BackendBase, backend_types
from caikit.core.modules import ModuleBase, ModuleConfig, ModuleSaver, module
from caikit.interfaces.nlp.data_model import (
    GeneratedTextResult,
    GeneratedTextStreamResult,
    TokenizationResults,
)
from caikit.interfaces.nlp.tasks import TextGenerationTask, TokenizationTask
from caikit.interfaces.runtime.data_model import RuntimeServerContextType
from caikit_tgis_backend import TGISBackend
import alog

# Local
from ...data_model import ExponentialDecayLengthPenalty
from ...resources.pretrained_model import (
    HFAutoCausalLM,
    HFAutoSeq2SeqLM,
    PretrainedModelBase,
)
from ...toolkit.text_generation.tgis_utils import (
    GENERATE_FUNCTION_TGIS_ARGS,
    TGISGenerationClient,
)
from .text_generation_local import TextGeneration

log = alog.use_channel("TXT_GEN")
error = error_handler.get(log)

# pylint: disable=too-many-instance-attributes


@module(
    backend_type=TGISBackend.backend_type,
    base_module=TextGeneration,
    tasks=[TextGenerationTask, TokenizationTask],
)
class TextGenerationTGIS(ModuleBase):
    """Module to provide text generation capabilities"""

    SUPPORTED_LOAD_BACKENDS = [TGISBackend.backend_type, backend_types.LOCAL]

    supported_resources = [HFAutoCausalLM, HFAutoSeq2SeqLM]

    def __init__(
        self,
        model_name: str,
        model: Optional[PretrainedModelBase] = None,
        bos_token: Optional[str] = None,
        sep_token: Optional[str] = None,
        eos_token: Optional[str] = None,
        pad_token: Optional[str] = None,
        tgis_backend: Optional[TGISBackend] = None,
    ):
        super().__init__()

        error.type_check("<NLP00609194E>", str, allow_none=True, bos_token=bos_token)
        error.type_check("<NLP72469403E>", str, allow_none=True, sep_token=sep_token)
        error.type_check("<NLP48137045E>", str, allow_none=True, eos_token=eos_token)
        error.type_check("<NLP53511308E>", str, allow_none=True, pad_token=pad_token)
        self.model = model
        self.model_name = model_name

        # Set _model_loaded as False by default. This will only get set to True if
        # we enable the tgis_backend and we are able to fetch the client successfully.
        self._model_loaded = False
        if tgis_backend:
            self.tgis_backend = tgis_backend

        self._tgis_backend = tgis_backend
        self._bos_token = bos_token
        self._sep_token = sep_token
        self._eos_token = eos_token
        self._pad_token = pad_token

    def __del__(self):
        # nothing to unload if we didn't finish loading
        if self._model_loaded and self._tgis_backend:
            self._tgis_backend.unload_model(self.model_name)

    @cached_property
    def _client(self):
        # Lazily configure/create the internal tgis backend client
        if self._tgis_backend:
            return self._tgis_backend.get_client(self.model_name)

    @cached_property
    def tgis_generation_client(self):
        # Lazily create the generation client
        # This in turn calls self._client which also lazily gets the tgis backend client
        return TGISGenerationClient(
            self.model_name, self._eos_token, self._client, self.PRODUCER_ID
        )

    @classmethod
    def bootstrap(cls, model_path: str, load_backend: Union[BackendBase, None] = None):
        """Function to bootstrap a pre-trained transformers model and
        get a caikit text-generation 'model'.

        Args:
            base_model_path: str
                Path to transformers model
                NOTE: Model path needs to contain tokenizer as well
            load_backend: BackendBase
                Backend object to be used to run inference with.
                NOTE: this is required for inferencing. It is
                made optional to support the model conversion use-case
        Returns:
            caikit_nlp.blocks.text_generation.TextGeneration
                Object of TextGeneration class (model)
        """

        text_generation_inst = TextGeneration.bootstrap(model_path)
        bos_token = text_generation_inst.model._tokenizer.bos_token
        sep_token = text_generation_inst.model._tokenizer.sep_token
        eos_token = text_generation_inst.model._tokenizer.eos_token or None
        pad_token = text_generation_inst.model._tokenizer.pad_token

        return cls(
            text_generation_inst.model_name,
            text_generation_inst.model,
            bos_token=bos_token,
            sep_token=sep_token,
            eos_token=eos_token,
            pad_token=pad_token,
            tgis_backend=load_backend,
        )

    @classmethod
    def load(cls, model_path: str, load_backend: BackendBase) -> "TextGeneration":
        """Function to load text-generation model. Note, this only loads
        "remote" style model, i.e the cakit-model that doesn't
        necessarily required to have actual artifacts in it
        and thus only saves them in "remote" format.

        Args:
            model_path: str
                Path to the model to be loaded.
            load_backend: BackendBase
                Backend object to be used to run inference with.
        Returns:
            TextGeneration
                Instance of this class built from the on disk model.
        """
        error.type_check("<NLP03521359E>", TGISBackend, load_backend=load_backend)

        config = ModuleConfig.load(model_path)
        tgis_backend = config.tgis_backend or load_backend
        artifacts_path = config.artifact_path
        if artifacts_path:
            model_name = os.path.join(model_path, artifacts_path)
            error.dir_check("<NLP01983374E>", model_name)
            log.debug("Loading with on-disk artifacts: %s", model_name)
        else:
            model_name = config.model_name
            error.type_check("<NLP90686335E>", str, model_name=model_name)
            log.debug("Loading with model name: %s", model_name)
        return cls(
            model_name,
            bos_token=config.bos_token,
            sep_token=config.sep_token,
            eos_token=config.eos_token,
            pad_token=config.pad_token,
            tgis_backend=tgis_backend,
        )

    def save(self, model_path: str):
        """Export the config for this model.
        This saves the model in "remote" style
        and does not store the actual model artifacts
        along with the caikit-model.

        model_path: str
            Path to which we should write our model.
        """
        # pylint: disable=duplicate-code
        saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with saver:
            saver.update_config(
                {
                    "model_name": self.model_name,
                    "bos_token": self._bos_token,
                    "sep_token": self._sep_token,
                    "eos_token": self._eos_token,
                    "pad_token": self._pad_token,
                }
            )

    # pylint: disable=duplicate-code
    @TextGenerationTask.taskmethod(context_arg="context")
    def run(
        self,
        text: str,
        max_new_tokens: Optional[int] = 20,
        min_new_tokens: Optional[int] = 0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        exponential_decay_length_penalty: Optional[
            Union[Tuple[int, float], ExponentialDecayLengthPenalty]
        ] = None,
        stop_sequences: Optional[List[str]] = None,
        seed: Optional[np.uint64] = None,
        preserve_input_text: bool = False,
        input_tokens: bool = False,
        generated_tokens: bool = True,
        token_logprobs: bool = True,
        token_ranks: bool = True,
        include_stop_sequence: Optional[bool] = True,
        context: Optional[RuntimeServerContextType] = None,
    ) -> GeneratedTextResult:
        f"""Run inference against the model running in TGIS.

        Args:
           {GENERATE_FUNCTION_TGIS_ARGS}
        Returns:
            GeneratedTextResult
                Generated text result produced by TGIS.
        """
        self._register_model_connection_with_context(context)

        if self._model_loaded:
            return self.tgis_generation_client.unary_generate(
                text=text,
                preserve_input_text=preserve_input_text,
                input_tokens=input_tokens,
                generated_tokens=generated_tokens,
                token_logprobs=token_logprobs,
                token_ranks=token_ranks,
                include_stop_sequence=include_stop_sequence,
                max_new_tokens=max_new_tokens,
                min_new_tokens=min_new_tokens,
                truncate_input_tokens=truncate_input_tokens,
                decoding_method=decoding_method,
                top_k=top_k,
                top_p=top_p,
                typical_p=typical_p,
                temperature=temperature,
                seed=seed,
                repetition_penalty=repetition_penalty,
                max_time=max_time,
                exponential_decay_length_penalty=exponential_decay_length_penalty,
                stop_sequences=stop_sequences,
            )

    @TextGenerationTask.taskmethod(output_streaming=True, context_arg="context")
    def run_stream_out(
        self,
        text: str,
        max_new_tokens: Optional[int] = 20,
        min_new_tokens: Optional[int] = 0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        exponential_decay_length_penalty: Optional[
            Union[Tuple[int, float], ExponentialDecayLengthPenalty]
        ] = None,
        stop_sequences: Optional[List[str]] = None,
        seed: Optional[np.uint64] = None,
        preserve_input_text: bool = False,
        input_tokens: bool = False,
        generated_tokens: bool = True,
        token_logprobs: bool = True,
        token_ranks: bool = True,
        include_stop_sequence: Optional[bool] = True,
        context: Optional[RuntimeServerContextType] = None,
    ) -> Iterable[GeneratedTextStreamResult]:
        f"""Run output stream inferencing for text generation module.

        Args:
            {GENERATE_FUNCTION_TGIS_ARGS}
        Returns:
            Iterable[GeneratedTextStreamResult]
        """
        self._register_model_connection_with_context(context)

        if self._model_loaded:
            return self.tgis_generation_client.stream_generate(
                text=text,
                preserve_input_text=preserve_input_text,
                input_tokens=input_tokens,
                generated_tokens=generated_tokens,
                token_logprobs=token_logprobs,
                token_ranks=token_ranks,
                include_stop_sequence=include_stop_sequence,
                max_new_tokens=max_new_tokens,
                min_new_tokens=min_new_tokens,
                truncate_input_tokens=truncate_input_tokens,
                decoding_method=decoding_method,
                top_k=top_k,
                top_p=top_p,
                typical_p=typical_p,
                temperature=temperature,
                seed=seed,
                repetition_penalty=repetition_penalty,
                max_time=max_time,
                exponential_decay_length_penalty=exponential_decay_length_penalty,
                stop_sequences=stop_sequences,
            )

    @TokenizationTask.taskmethod(context_arg="context")
    def run_tokenizer(
        self,
        text: str,
        context: Optional[RuntimeServerContextType] = None,
    ) -> TokenizationResults:
        """Run tokenization task against the model running in TGIS.

        Args:
           text: str
                Text to tokenize
        Returns:
            TokenizationResults
                The token count
        """
        self._register_model_connection_with_context(context)

        if self._model_loaded:
            return self.tgis_generation_client.unary_tokenize(
                text=text,
            )

    def _register_model_connection_with_context(
        self, context: Optional[RuntimeServerContextType]
    ):
        """
        Register a remote model connection with the configured TGISBackend if there is
        a context override provided.
        """
        if self._tgis_backend:
            self._tgis_backend.handle_runtime_context(self.model_name, context)
            self._model_loaded = True



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for tgis_utils
"""

# Standard
from typing import Iterable, Optional, Type

# Third Party
import fastapi
import grpc
import grpc._channel
import pytest

# First Party
from caikit.core.data_model import ProducerId
from caikit.core.exceptions.caikit_core_exception import CaikitCoreException
from caikit.interfaces.runtime.data_model import RuntimeServerContextType
from caikit_tgis_backend.protobufs import generation_pb2

# Local
from caikit_nlp.toolkit.text_generation import tgis_utils
from tests.fixtures import TestServicerContext

## Helpers #####################################################################


class MockTgisClient:
    """Mock of a TGIS client that doesn't actually call anything"""

    def __init__(
        self,
        status_code: Optional[grpc.StatusCode],
        error_message: str = "Yikes",
    ):
        self._status_code = status_code
        self._error_message = error_message

    def _maybe_raise(self, error_type: Type[grpc.RpcError], *args):
        if self._status_code not in [None, grpc.StatusCode.OK]:
            raise error_type(
                grpc._channel._RPCState(
                    [], [], [], code=self._status_code, details=self._error_message
                ),
                *args,
            )

    def Generate(
        self, request: generation_pb2.BatchedGenerationRequest, **kwargs
    ) -> generation_pb2.BatchedGenerationResponse:
        self._maybe_raise(grpc._channel._InactiveRpcError)
        return generation_pb2.BatchedGenerationResponse()

    def GenerateStream(
        self, request: generation_pb2.SingleGenerationRequest, **kwargs
    ) -> Iterable[generation_pb2.GenerationResponse]:
        self._maybe_raise(grpc._channel._MultiThreadedRendezvous, None, None, None)
        yield generation_pb2.GenerationResponse()

    def Tokenize(
        self, request: generation_pb2.BatchedTokenizeRequest, **kwargs
    ) -> generation_pb2.BatchedTokenizeResponse:
        self._maybe_raise(grpc._channel._InactiveRpcError)
        return generation_pb2.BatchedTokenizeResponse()


## TGISGenerationClient ########################################################


@pytest.mark.parametrize(
    "status_code",
    [code for code in grpc.StatusCode if code != grpc.StatusCode.OK],
)
@pytest.mark.parametrize(
    "method", ["unary_generate", "stream_generate", "unary_tokenize"]
)
def test_TGISGenerationClient_rpc_errors(status_code, method):
    """Test that raised errors in downstream RPCs are converted to
    CaikitCoreException correctly
    """
    tgis_client = MockTgisClient(status_code)
    gen_client = tgis_utils.TGISGenerationClient(
        "foo",
        "bar",
        tgis_client,
        ProducerId("foobar"),
    )
    with pytest.raises(CaikitCoreException) as context:
        kwargs = (
            dict(
                preserve_input_text=True,
                input_tokens=True,
                generated_tokens=True,
                token_logprobs=True,
                token_ranks=True,
                max_new_tokens=20,
                min_new_tokens=20,
                truncate_input_tokens=True,
                decoding_method="GREEDY",
                top_k=None,
                top_p=None,
                typical_p=None,
                temperature=None,
                seed=None,
                repetition_penalty=0.5,
                max_time=None,
                exponential_decay_length_penalty=None,
                stop_sequences=["asdf"],
                include_stop_sequence=True,
            )
            if method.endswith("_generate")
            else dict()
        )
        res = getattr(gen_client, method)(text="foobar", **kwargs)
        if method.startswith("stream_"):
            next(res)

    assert (
        context.value.status_code == tgis_utils.GRPC_TO_CAIKIT_CORE_STATUS[status_code]
    )
    rpc_err = context.value.__context__
    assert isinstance(rpc_err, grpc.RpcError)


# NOTE: This test is preserved in caikit-nlp despite being duplicated in
# caikit-tgis-backend so that we guarantee that the functionality is accessible
# in a version-compatible way here.
@pytest.mark.parametrize(
    argnames=["context", "route_info"],
    argvalues=[
        (
            fastapi.Request(
                {
                    "type": "http",
                    "headers": [
                        (tgis_utils.ROUTE_INFO_HEADER_KEY.encode(), b"sometext")
                    ],
                }
            ),
            "sometext",
        ),
        (
            fastapi.Request(
                {"type": "http", "headers": [(b"route-info", b"sometext")]}
            ),
            None,
        ),
        (
            TestServicerContext({tgis_utils.ROUTE_INFO_HEADER_KEY: "sometext"}),
            "sometext",
        ),
        (
            TestServicerContext({"route-info": "sometext"}),
            None,
        ),
        ("should raise ValueError", None),
        (None, None),
        # Uncertain how to create a grpc.ServicerContext object
    ],
)
def test_get_route_info(context: RuntimeServerContextType, route_info: Optional[str]):
    if not isinstance(context, (fastapi.Request, grpc.ServicerContext, type(None))):
        with pytest.raises(TypeError):
            tgis_utils.get_route_info(context)
    else:
        actual_route_info = tgis_utils.get_route_info(context)
        assert actual_route_info == route_info



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This sets up global test configs when pytest starts
"""

# Standard
import os

# First Party
import alog

# Configure logging from the environment
alog.configure(
    default_level=os.environ.get("LOG_LEVEL", "off"),
    filters=os.environ.get("LOG_FILTERS", "urllib3:off"),
    thread_id=os.environ.get("LOG_THREAD_ID", "") == "true",
)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains toolkit functionality for huggingface Trainer"""
# Standard
from datetime import datetime

# Third Party
import torch

# First Party
from caikit import get_config
from caikit.core.data_model import DataStream
from caikit.core.exceptions import error_handler
import alog

log = alog.use_channel("TRNR_UTILS")
error = error_handler.get(log)


def validate_training_data(train_stream: DataStream, model_name: str, module_id: str):

    global_default = get_config().training_data_limit.__default__
    module_default = (
        get_config()
        .training_data_limit.get(module_id, {})
        .get("__default__", global_default)
    )

    max_num_examples = (
        get_config()
        .training_data_limit.get(module_id, {})
        .get(model_name, module_default)
    )

    if max_num_examples > 0:
        train_stream_size = len(train_stream)
        error.value_check(
            "<NLP77627434E>",
            train_stream_size <= max_num_examples,
            "Number of examples ({}) exceeds the maximum number of examples allowed "
            "({}) for this model",
            train_stream_size,
            max_num_examples,
        )


def log_step(state, logs):
    if state.epoch is not None:
        logs["epoch"] = round(state.epoch, 2)

    # Get Rank
    if torch.distributed.is_initialized():
        rank = torch.distributed.get_rank()
    else:
        rank = 0

    if "loss" in logs:
        if state.epoch is not None:
            logs["epoch"] = round(state.epoch, 2)

        log.debug(
            "process rank: {} loss: {} step: {}".format(
                rank, float(logs["loss"]), state.global_step
            )
        )
        output = {
            "epoch": float(logs["epoch"]),
            "step": state.global_step,
            "value": float(logs["loss"]),
            "timestamp": datetime.isoformat(datetime.now()),
        }
        state.log_history.append(output)
    else:
        output = {**logs, **{"step": state.global_step}}
        state.log_history.append(output)

    return state



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from pathlib import Path

# First Party
import alog
import caikit

log = alog.use_channel("CONFIG_INIT")

# The name for an extension is simply the name of the directory containing its config dir.
extension_name = Path(__file__).parent.parent.name

MODEL_MANAGER = caikit.core.MODEL_MANAGER

extract = MODEL_MANAGER.extract
load = MODEL_MANAGER.load
resolve_and_load = MODEL_MANAGER.resolve_and_load



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module that provides capability to split documents into sentences via regex"""

# Standard
import os
import re

# First Party
from caikit.core.exceptions import error_handler
from caikit.core.modules import ModuleBase, ModuleConfig, ModuleSaver, module
from caikit.interfaces.nlp.data_model import Token, TokenizationResults
from caikit.interfaces.nlp.tasks import TokenizationTask
import alog

log = alog.use_channel("RGX_SNT_SPLT")
error = error_handler.get(log)


@module(
    id="1e04e21b-7009-499e-abdd-41e5984c2e7d",
    name="Regex Sentence Splitter",
    version="0.1.0",
    task=TokenizationTask,
)
class RegexSentenceSplitter(ModuleBase):
    # pylint: disable=anomalous-backslash-in-string
    """Use python regexes to split document into sentences.

    Sample regex string:
        [^.!?\s][^.!?\n]*(?:[.!?](?!['\"]?\s|$)[^.!?]*)*[.!?]?['\"]?(?=\s|$)
    """

    def __init__(self, regex_str: str):
        """Construct a RegexSentenceSplitter object
        by compiling the input regex string into python regex object
        that can be used later on for detection.

        Args:
            regex_str: str
                String containing pattern that can be complied with python re
                module
        """
        super().__init__()
        error.type_check("<NLP48846517E>", str, regex_str=regex_str)
        self.regex_str = regex_str
        self.regex = re.compile(self.regex_str)

    @classmethod
    def bootstrap(cls, regex_str) -> "RegexSentenceSplitter":
        """Bootstrap a a RegexSentenceSplitter object

        Args:
            regex_str: str
                String containing pattern that can be complied with python re
                module
        """
        return cls(regex_str)

    def save(self, model_path: str):
        """Save model in target path

        Args:
            model_path: str
                Path to store model artifact(s)
        """
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            config_options = {"regex_str": self.regex_str}
            module_saver.update_config(config_options)

    @classmethod
    def load(cls, model_path: str) -> "RegexSentenceSplitter":
        """Load a regex sentence splitter model.

        Args:
            model_path: str
                Path to the model to be loaded.

        Returns:
            RegexSentenceSplitter
                Instance of this class built from the on disk model.
        """
        config = ModuleConfig.load(os.path.abspath(model_path))
        return cls(regex_str=config.regex_str)

    def run(self, text: str) -> TokenizationResults:
        """Run sentence splitting regex on input text.

        Args:
            text: str
                Document to run sentence splitting on.

        Returns:
            TokenizationResults
                TokenizationResults object containing tokens where each token
                corresponds to a detected sentence.
        """

        error.type_check("<NLP38553904E>", str, text=text)

        matches = self.regex.finditer(text)
        tokens = []
        for match in matches:
            token = Token(start=match.start(), end=match.end(), text=match.group())
            tokens.append(token)

        return TokenizationResults(results=tokens)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data structures for text generation representations
"""
# Standard
from enum import Enum
from typing import List

# First Party
from caikit.core import DataObjectBase

# First party
import alog
import caikit

log = alog.use_channel("DATAM")


class PromptOutputModelType(Enum):
    ENCODER = "ENCODER"
    DECODER = "DECODER"


@caikit.core.dataobject(package="caikit_data_model.caikit_nlp")
class GenerationTrainRecord(DataObjectBase):
    input: str
    output: str


@caikit.core.dataobject(package="caikit_data_model.caikit_nlp")
class TuningConfig(DataObjectBase):
    # If prompt_tuning_init_text is not provided, then random would be used
    # but since random is not supported currently, we want to keep text to be
    # required
    num_virtual_tokens: int
    # TODO: Move all _init_ params to separate object
    prompt_tuning_init_text: str
    prompt_tuning_init_method: str
    #    could be: `RANDOM`, `TEXT`, `ONLY_SOURCE_SHARED` and `AVERAGE_SOURCE`
    #
    prompt_tuning_init_source_model: str  # this maps to prompt_tuning_init_state_dict_path in MPT
    #     which is path pointing to the state dict of the model to be used for initialization
    #
    # token_dim: int # Optional - dimension of the virtual tokens.
    #
    output_model_types: List[str]
    # this replaces `num_transformer_submodules`
    #     option and can take values encoder, decoder. For each
    #     selected resource type will only provide certain possibilities,
    #     for example, causal-lm models will only provide decoder as option.
    #     If None provided, then we will use defaults for that model_type
    # num_transformer_submodules: int # Optional - The number of transformer submodules in the
    # base transformer model.
    #     1 for all decoder-only models and 2 for encoder-decoder models.
    #     If 1 is used for encoder-decoder models, the prompt will be used for the encoder only.
    #
    # num_attention_heads: int # Optional - The number of attention heads in the
    # base transformer model
    #
    # num_layers: int # Optional - The number of layers in the base transformer model
    #
    # encoder_hidden_size: int # Optional -  The hidden size of the prompt encoder.


@caikit.core.dataobject(package="caikit_data_model.caikit_nlp")
class ExponentialDecayLengthPenalty(DataObjectBase):
    start_index: int
    decay_factor: float



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a distributed backend implementation for leveraging the PEFT-trained
prompt vectors in TGIS generation requests.
"""

# Standard
from functools import cached_property
from typing import Iterable, List, Optional, Tuple, Union
import os

# Third Party
import numpy as np

# First Party
from caikit.config import get_config
from caikit.core import ModuleBase, ModuleConfig, ModuleSaver, modules
from caikit.core.exceptions import error_handler
from caikit.core.module_backends import BackendBase, backend_types
from caikit.interfaces.nlp.data_model import (
    GeneratedTextResult,
    GeneratedTextStreamResult,
    TokenizationResults,
)
from caikit.interfaces.nlp.tasks import TextGenerationTask, TokenizationTask
from caikit.interfaces.runtime.data_model import RuntimeServerContextType
from caikit_tgis_backend import TGISBackend
import alog

# Local
from ...data_model import ExponentialDecayLengthPenalty
from ...toolkit.text_generation.tgis_utils import (
    GENERATE_FUNCTION_TGIS_ARGS,
    TGISGenerationClient,
)
from ...toolkit.verbalizer_utils import render_verbalizer
from . import PeftPromptTuning

log = alog.use_channel("PEFT_PROMPT_REMOTE")
error = error_handler.get(log)


@modules.module(
    backend_type=TGISBackend.backend_type,
    base_module=PeftPromptTuning,
    tasks=[TextGenerationTask, TokenizationTask],
)
class PeftPromptTuningTGIS(ModuleBase):  # pylint: disable=too-many-instance-attributes
    SUPPORTED_LOAD_BACKENDS = [TGISBackend.backend_type, backend_types.LOCAL]
    ## Module Interface ##

    def __init__(
        self,
        base_model_name: str,
        prompt_cache_id: str,
        eos_token: str,
        verbalizer: str,
        enable_backend: bool = True,
        tgis_backend: Optional[TGISBackend] = None,
        prompt_artifacts: Optional[List[str]] = None,
    ) -> None:
        super().__init__()

        self._tgis_backend = tgis_backend
        if enable_backend:
            error.type_check(
                "<NLP33971947E>", TGISBackend, tgis_backend=self._tgis_backend
            )
            # get_client will also launch a local TGIS process and get the model
            # loaded when using the local TGIS backend

            # Tell the backend to load all of the available prompt files
            if prompt_artifacts:
                tgis_backend.load_prompt_artifacts(
                    base_model_name, prompt_cache_id, *prompt_artifacts
                )

        self.base_model_name = base_model_name
        self._prompt_cache_id = prompt_cache_id
        self.eos_token = eos_token
        self.verbalizer = verbalizer
        self.enable_backend = enable_backend

        self.tgis_generation_client = TGISGenerationClient(
            self.base_model_name,
            self.eos_token,
            self._client,
            self.PRODUCER_ID,
            self._prompt_cache_id,
        )

    def __del__(self):
        """Attempt to clean up the prompt cache on deletion"""
        if get_config().unload_tgis_prompt_artifacts:
            tgis_backend = getattr(self, "_tgis_backend", None)
            prompt_cache_id = getattr(self, "_prompt_cache_id", None)
            model_id = getattr(self, "base_model_name", None)
            if tgis_backend and prompt_cache_id and model_id:
                tgis_backend.unload_prompt_artifacts(model_id, prompt_cache_id)

    @cached_property
    def _client(self):
        # Configure the internal client
        # NOTE: This is made optional for the cases where we do not need to execute `.run` function
        # for example, bootstrapping a model to caikit format and saving.
        if self._tgis_backend:
            return self._tgis_backend.get_client(self.base_model_name)

    @classmethod
    def load(cls, model_path: str, load_backend: BackendBase) -> "PeftPromptTuningTGIS":
        """Load a TGIS Peft Prompt Tuning distributed module. Note that we do not
        leverage artifacts stored within the model here, and we assume that the
        prompt vector is already available at a place that the TGIS server can pick it
        up.

        Args:
            model_path: str
                Path to the model to be loaded.
            load_backend: BackendBase
                Backend object to be used to run inference with.
        Returns:
            PeftPromptTuningTGIS
                Instance of this class built from the on disk model.
        """
        error.type_check("<NLP85069377E>", TGISBackend, load_backend=load_backend)
        config = ModuleConfig.load(model_path)
        eos_token = config.eos_token
        verbalizer = config.verbalizer
        dir_name = os.path.basename(model_path)
        # NOTE: base_model_name is used as "model_id" when calling to TGIS backend
        base_model_name = config.get("base_model_name", "")
        prompt_cache_id = dir_name
        error.type_check("<NLP24633932E>", str, prompt_cache_id=prompt_cache_id)
        # NOTE: prompt model config stores a base_model_config
        # which can be used to validate if the prompt is tuned
        # for the model it is being used with. However,
        # we are currently not accessing or assuming the accessibilty of
        # base model, thus not validating.
        # NOTE: When we access base_model_config, we need to make sure
        # we convert make it valid json compatible dict (aka doesn't have non string keys)
        log.debug("Prompt ID: %s", prompt_cache_id)
        log.debug("TGIS model ID: %s", base_model_name)

        # Get all the valid prompt artifact files so they can be loaded after
        # the connection is established
        prompt_artifacts = [
            os.path.join(model_path, config.get(config_key))
            for config_key in [
                PeftPromptTuning._ENCODER_KEY.name,
                PeftPromptTuning._DECODER_KEY.name,
            ]
            if config.get(config_key)
        ]
        return cls(
            base_model_name,
            prompt_cache_id,
            eos_token,
            verbalizer,
            tgis_backend=load_backend,
            prompt_artifacts=prompt_artifacts,
        )

    def save(self, model_path: str):
        """Export the config for this model.

        model_path: str
            Path to which we should write our model.
        """
        saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with saver:
            saver.update_config(
                {
                    "base_model_name": self.base_model_name,
                    "prompt_cache_id": self._prompt_cache_id,
                    "eos_token": self.eos_token,
                    "verbalizer": self.verbalizer,
                }
            )

    # pylint: disable=duplicate-code
    @TextGenerationTask.taskmethod(context_arg="context")
    def run(
        self,
        text: str,
        max_new_tokens: Optional[int] = 20,
        min_new_tokens: Optional[int] = 0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        exponential_decay_length_penalty: Optional[
            Union[Tuple[int, float], ExponentialDecayLengthPenalty]
        ] = None,
        stop_sequences: Optional[List[str]] = None,
        seed: Optional[np.uint64] = None,
        preserve_input_text: bool = False,
        input_tokens: bool = False,
        generated_tokens: bool = True,
        token_logprobs: bool = True,
        token_ranks: bool = True,
        include_stop_sequence: Optional[bool] = True,
        context: Optional[RuntimeServerContextType] = None,
    ) -> GeneratedTextResult:
        f"""Run inference against the model running in TGIS.

        Args:
           {GENERATE_FUNCTION_TGIS_ARGS}
        Returns:
            GeneratedTextResult
                Generated text result produced by TGIS.
        """

        error.value_check(
            "<NLP87360638E>",
            self.enable_backend,
            "Backend must be configured and loaded with this module before executing `run` call.",
        )
        self._register_model_connection_with_context(context)

        verbalized_text = render_verbalizer(self.verbalizer, {"input": text})
        return self.tgis_generation_client.unary_generate(
            text=verbalized_text,
            preserve_input_text=preserve_input_text,
            input_tokens=input_tokens,
            generated_tokens=generated_tokens,
            token_logprobs=token_logprobs,
            token_ranks=token_ranks,
            include_stop_sequence=include_stop_sequence,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

    @TextGenerationTask.taskmethod(output_streaming=True, context_arg="context")
    def run_stream_out(
        self,
        text: str,
        max_new_tokens: Optional[int] = 20,
        min_new_tokens: Optional[int] = 0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        exponential_decay_length_penalty: Optional[
            Union[Tuple[int, float], ExponentialDecayLengthPenalty]
        ] = None,
        stop_sequences: Optional[List[str]] = None,
        seed: Optional[np.uint64] = None,
        preserve_input_text: bool = False,
        input_tokens: bool = False,
        generated_tokens: bool = True,
        token_logprobs: bool = True,
        token_ranks: bool = True,
        include_stop_sequence: Optional[bool] = True,
        context: Optional[RuntimeServerContextType] = None,
    ) -> Iterable[GeneratedTextStreamResult]:
        f"""Run output stream inferencing against the model running in TGIS

        Args:
            {GENERATE_FUNCTION_TGIS_ARGS}
        Returns:
            Iterable[GeneratedTextStreamResult]
        """

        error.value_check(
            "<NLP62995899E>",
            self.enable_backend,
            "Backend must be configured and loaded with this module \
            before executing `run_stream_out` call.",
        )

        self._register_model_connection_with_context(context)

        verbalized_text = render_verbalizer(self.verbalizer, {"input": text})
        return self.tgis_generation_client.stream_generate(
            text=verbalized_text,
            preserve_input_text=preserve_input_text,
            input_tokens=input_tokens,
            generated_tokens=generated_tokens,
            token_logprobs=token_logprobs,
            token_ranks=token_ranks,
            include_stop_sequence=include_stop_sequence,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

    @TokenizationTask.taskmethod(context_arg="context")
    def run_tokenizer(
        self,
        text: str,
        context: Optional[RuntimeServerContextType] = None,
    ) -> TokenizationResults:
        """Run tokenization task against the model running in TGIS.

        Args:
            text: str
                Text to tokenize
        Returns:
            TokenizationResults
                The token count
        """

        self._register_model_connection_with_context(context)

        return self.tgis_generation_client.unary_tokenize(
            text=text,
        )

    def _register_model_connection_with_context(
        self, context: Optional[RuntimeServerContextType]
    ):
        """
        Register a remote model connection with the configured TGISBackend if there is
        a context override provided.
        """
        if self._tgis_backend:
            self._tgis_backend.handle_runtime_context(self.base_model_name, context)
            self._model_loaded = True



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Third Party
import pytest

# First Party
from caikit.interfaces.nlp.data_model import ClassificationTrainRecord

# Local
from caikit_nlp.data_model import GenerationTrainRecord
from caikit_nlp.toolkit.task_specific_utils import convert_to_generation_record


def test_convert_classification_train_record_to_generation_record():
    classification_train_record = ClassificationTrainRecord(
        text="foo bar", labels=["label1"]
    )
    generated_train = convert_to_generation_record(classification_train_record)
    assert isinstance(generated_train, GenerationTrainRecord)
    assert generated_train.input == "foo bar"
    assert generated_train.output == "label1"


def test_convert_generation_record_to_generation_record():
    generation_train_record = GenerationTrainRecord(input="foo bar", output="label1")
    generated_train = convert_to_generation_record(generation_train_record)
    assert isinstance(generated_train, GenerationTrainRecord)
    assert generated_train.input == generation_train_record.input
    assert generated_train.output == generation_train_record.output


# When we support integer labels
# def test_convert_classification_train_record_to_generation_record_numeric_labels():
#     classification_train_record = dm.ClassificationTrainRecord(
#         text="foo bar", labels=[1]
#     )
#     generated_train = convert_to_generation_record(classification_train_record)
#     assert isinstance(generated_train, dm.GenerationTrainRecord)
#     assert generated_train.input == classification_train_record.text
#     assert generated_train.output == "1"


def test_convert_to_generation_record_gives_error_with_unsupported_type():
    string_record = "test record"
    with pytest.raises(TypeError):
        convert_to_generation_record(string_record)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module returns span classification output by splitting
text into spans and returning classifications for each span. Span
classifications can be filtered by score threshold and label(s).
At this time this module is only designed for inference"""

# Standard
from typing import Iterable, List, Optional, Union
import itertools
import os

# First Party
from caikit.core.exceptions import error_handler
from caikit.core.modules import (
    ModuleBase,
    ModuleConfig,
    ModuleLoader,
    ModuleSaver,
    module,
)
from caikit.interfaces.nlp.data_model import (
    TokenClassificationResult,
    TokenClassificationResults,
    TokenClassificationStreamResult,
)
from caikit.interfaces.nlp.tasks import (
    TextClassificationTask,
    TokenClassificationTask,
    TokenizationTask,
)
import alog

log = alog.use_channel("FILT_SPAN")
error = error_handler.get(log)

# Tasks allowed for the module to be provided for classifier
ALLOWED_TASKS = [TextClassificationTask, TokenClassificationTask]


@module(
    id="42a7d920-8b7e-4e1f-81fb-8ab851a80c99",
    name="Filtered span classification",
    version="0.1.0",
    task=TokenClassificationTask,
)
class FilteredSpanClassification(ModuleBase):

    ################################ Constructor #################################################

    def __init__(
        self,
        lang: str,
        tokenizer: ModuleBase,
        classifier: ModuleBase,
        default_threshold: float,
        labels_to_output: List[str] = None,
    ):
        """Construct a filtered span classification object
        from a tokenizer and sequence classifier

        Args:
            lang: str
                2 letter language code
            tokenizer: ModuleBase
                Tokenizer that returns TokenizationResults
            classifier: ModuleBase
                Classification model instance returning Classification or
                TokenClassificationResult output on .run
            default_threshold: float
                Default threshold for scores
            labels_to_output: List[str]
                (Optional) Select labels to output, if None all labels will be returned
        """
        super().__init__()
        error.type_check("<NLP12578168E>", str, lang=lang)
        error.type_check("<NLP79642537E>", ModuleBase, tokenizer=tokenizer)
        error.value_check(
            "<NLP42736791E>",
            TokenizationTask in type(tokenizer).tasks,
            "tokenizer does not implement TokenizationTask",
        )
        error.type_check(
            "<NLP35742128E>",
            ModuleBase,
            classifier=classifier,
        )
        error.type_check("<NLP63802045E>", float, default_threshold=default_threshold)
        error.type_check_all(
            "<NLP71653678E>", str, allow_none=True, labels_to_output=labels_to_output
        )
        classification_tasks = type(classifier).tasks
        tasks_intersection = [i for i in classification_tasks if i in ALLOWED_TASKS]
        error.value_check(
            "<NLP41319814E>",
            any(tasks_intersection),
            f"classifier does not implement one of required tasks: {ALLOWED_TASKS}",
        )
        error.value_check(
            "<NLP41319815E>",
            len(tasks_intersection) == 1,
            f"classifier should implement only one task in: {ALLOWED_TASKS}",
        )
        self.lang = lang
        self.tokenizer = tokenizer
        self.classifier = classifier
        self.default_threshold = default_threshold
        self.labels_to_output = labels_to_output
        self.classification_task = tasks_intersection[0]

    ################################## API functions #############################################

    @TokenClassificationTask.taskmethod()
    def run(
        self, text: str, threshold: Optional[Union[float, int]] = None
    ) -> TokenClassificationResults:
        """Run classification on text split into spans. Returns results
        based on score threshold for labels that are to be outputted

        Args:
            text: str
                Document to run classification on
            threshold: float | int
                (Optional) Threshold based on which to return score results

        Returns:
            TokenClassificationResults
        """
        error.type_check("<NLP82129006E>", str, text=text)
        error.type_check(
            "<NLP01414077E>", float, int, allow_none=True, threshold=threshold
        )

        if threshold is None:
            threshold = self.default_threshold
        if not text:
            # Allow empty text case to fall through - some tokenizers or
            # classifiers may error on this
            return TokenClassificationResults(results=[])

        token_classification_results = []
        if self.classification_task == TextClassificationTask:
            # Split document into spans
            span_list = self.tokenizer.run(text).results
            text_list = [span.text for span in span_list]
        else:
            # TokenClassificationTask classifiers would hold span info
            # so we do not need to span split again
            text_list = [text]
        # Run each span through the classifier and determine based
        # on threshold and labels_to_output what results should be returned
        classification_results = self.classifier.run_batch(text_list)
        for idx, classification_result in enumerate(classification_results):
            # Each classification result is list of classifications
            # for that particular text example
            for classification in classification_result.results:
                if self.classification_task == TextClassificationTask:
                    label = classification.label
                    span = span_list[idx]
                    start = span.start
                    end = span.end
                    word = span.text
                else:
                    label = classification.entity
                    start = classification.start
                    end = classification.end
                    word = classification.word
                if classification.score >= threshold:
                    if not self.labels_to_output or (
                        self.labels_to_output and label in self.labels_to_output
                    ):
                        token_classification = TokenClassificationResult(
                            start=start,
                            end=end,
                            word=word,
                            entity=label,
                            score=classification.score,
                        )
                        token_classification_results.append(token_classification)
        return TokenClassificationResults(results=token_classification_results)

    @TokenClassificationTask.taskmethod(input_streaming=True, output_streaming=True)
    def run_bidi_stream(
        self, text_stream: Iterable[str], threshold: Optional[Union[float, int]] = None
    ) -> Iterable[TokenClassificationStreamResult]:
        """Run bi-directional streaming inferencing for this module.
        Run classification on text split into spans. Returns results
        based on score threshold for labels that are to be outputted

        Args:
            text_stream: Iterable[str]
                Text stream to run classification on
            threshold: float | int
                (Optional) Threshold based on which to return score results

        Returns:
            Iterable[TokenClassificationStreamResult]
        """
        error.type_check(
            "<NLP96166348E>", float, int, allow_none=True, threshold=threshold
        )
        # TODO: For optimization implement window based approach.
        if threshold is None:
            threshold = self.default_threshold

        # Avoid length check here since it can be time consuming to iterate through stream
        # Tee stream to 2 - one to check emptiness, one for full iteration + analysis
        text_streams = itertools.tee(text_stream, 2)
        try:
            next(text_streams[0])
        except StopIteration:
            # Types on the stream are checked later on iteration
            # Allow empty text case to fall through - some tokenizers or
            # classifiers may error on this
            yield TokenClassificationStreamResult(results=[], processed_index=0)

        for span_output in self._stream_span_output(text_streams[1]):
            classification_result = self.classifier.run(span_output.text)
            results_to_end_of_span = False
            for classification in classification_result.results:
                if self.classification_task == TextClassificationTask:
                    label = classification.label
                    start = span_output.start
                    end = span_output.end
                    word = span_output.text
                else:
                    label = classification.entity
                    start = classification.start + span_output.start
                    end = classification.end + span_output.start
                    word = classification.word

                if classification.score >= threshold:
                    if not self.labels_to_output or (
                        self.labels_to_output and label in self.labels_to_output
                    ):
                        # Need to add offset to track actual place of spans within a stream,
                        # as the span splitting will be expected to stream and detect spans
                        yield TokenClassificationStreamResult(
                            results=[
                                TokenClassificationResult(
                                    start=start,
                                    end=end,
                                    word=word,
                                    entity=label,
                                    score=classification.score,
                                )
                            ],
                            processed_index=end,
                        )
                        if end == span_output.end:
                            results_to_end_of_span = True

            if not results_to_end_of_span:
                yield TokenClassificationStreamResult(
                    results=[], processed_index=span_output.end
                )

    def save(self, model_path: str):
        """Save model in target path

        Args:
            model_path: str
                Path to store model artifact(s)
        """
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )

        with module_saver:
            module_saver.save_module(self.tokenizer, "tokenizer")
            module_saver.save_module(self.classifier, "classification")
            config_options = {
                "language": self.lang,
                "default_threshold": self.default_threshold,
                "labels_to_output": self.labels_to_output,
            }
            module_saver.update_config(config_options)

    @classmethod
    def load(cls, model_path: str) -> "FilteredSpanClassification":
        """Load a filtered span classification model.

        Args:
            model_path: str
                Path to the model to be loaded.

        Returns:
            FilteredSpanClassification
                Instance of this class built from the on disk model.
        """
        config = ModuleConfig.load(os.path.abspath(model_path))
        loader = ModuleLoader(model_path)
        tokenizer = loader.load_module("tokenizer")
        classifier = loader.load_module("classification")
        return cls(
            tokenizer=tokenizer,
            classifier=classifier,
            lang=config.language,
            default_threshold=config.default_threshold,
            labels_to_output=config.labels_to_output,
        )

    @classmethod
    def bootstrap(
        cls,
        lang: str,
        tokenizer: ModuleBase,
        classifier: ModuleBase,
        default_threshold: float,
        labels_to_output: List[str] = None,
    ) -> "FilteredSpanClassification":
        """Bootstrap a FilteredSpanClassification instance

        Args:
            lang: str
                2 letter language code
            tokenizer: ModuleBase
                Tokenizer that returns TokenizationResults
            classifier: ModuleBase
                Classification model instance returning Classification or
                TokenClassificationResult output on .run
            default_threshold: float
                Default threshold for scores
            labels_to_output: List[str]
                (Optional) Select labels to output, if None all labels will be returned
        """
        # Basically just wrap the constructor currently
        return cls(
            lang=lang,
            tokenizer=tokenizer,
            classifier=classifier,
            default_threshold=default_threshold,
            labels_to_output=labels_to_output,
        )

    ################################## Private functions ##########################################

    def _stream_span_output(self, text_stream):
        """Function to yield span output from input text stream"""
        stream_accumulator = ""
        detected_spans = None
        span_start_offset = 0

        def __update_spans(token):
            # Check if the starting offset for the token is greater than
            # span_start_offset already, in which case, we need not
            # update the span
            if token.start < span_start_offset:
                # This is indicating that starting offset of token is off as we expect the token
                # to start at span_start_offset+1. So we need to recalibrate the token offsets
                # and have them start at span_start_offset. This means we need to know
                # the length of the token to manipulate the token.end, which we do by
                # subtracting end - start
                original_start = token.start
                token.start = span_start_offset
                token.end = span_start_offset + (token.end - original_start)
            return token

        for text in text_stream:
            error.type_check("<NLP38357927E>", str, text=text)
            stream_accumulator += text
            # In order to avoid processing all of the spans again, we only
            # send out the spans that are not yet finalized in detected_spans
            detected_spans = self.tokenizer.run(
                stream_accumulator[span_start_offset:]
            ).results

            if len(detected_spans) > 1:

                # we have detected more than 1 sentence
                # return 1st sentence
                new_span = detected_spans.pop(0)

                new_span = __update_spans(new_span)

                # we have detected new sentence, return the new sentence
                yield new_span

                # since we only send out part of the text, this means
                # the spans returned by the tokenizer will all now
                # start with last token offset + 1. So we store the last processed count
                # as the starting point for the subsequent one
                span_start_offset = new_span.end + 1

        # Return remaining sentence(s)
        if detected_spans and len(detected_spans) > 0:
            for detected_span in detected_spans:
                new_span = __update_spans(detected_span)
                yield new_span



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The TGISAutoFinder implements the ModelFinder interface to provide automatic
discovery of text-generation models that can be auto-configured to run against
a remote TGIS model.
"""
# Standard
from typing import Optional

# First Party
from caikit.core import MODEL_MANAGER, error_handler
from caikit.core.model_management import ModelFinderBase, model_finder_factory
from caikit.core.modules import ModuleConfig
from caikit_tgis_backend import TGISBackend
import aconfig
import alog

# Local
from ..modules.text_generation import TextGenerationTGIS

log = alog.use_channel("TGIS_FND")
error = error_handler.get(log)


class TGISAutoFinder(ModelFinderBase):
    __doc__ = __doc__

    name = "TGIS-AUTO"

    # Constants for the keys of the config blob
    _LOCAL_INITIALIZER_NAME_KEY = "local_initializer_name"
    _TGIS_BACKEND_PRIORITY_KEY = "tgis_backend_priority"

    def __init__(self, config: aconfig.Config, instance_name: str = ""):
        """Initialize from the model finder factory config

        Config schema:

        local_initializer_name:
            type: string
            default: "default"
            description: The name within the initializers config for the LOCAL
                initializer that will hold the tgis backend to use

        tgis_backend_priority:
            type: integer
            description: Index within the backend_priority list for the TGIS
                backend to use. If not set, the first TGIS backend found will be
                used.

        Args:
            config (aconfig.Config): The configuration blob from caikit's
                model_management factory construction
            instance_name (str): The name of this finder instance
        """
        local_initializer_name = config.get(self._LOCAL_INITIALIZER_NAME_KEY, "default")
        tgis_backend_priority = config.get(self._TGIS_BACKEND_PRIORITY_KEY)
        error.type_check(
            "<NLP97312902E>", str, local_initializer_name=local_initializer_name
        )
        error.type_check(
            "<NLP97312903E>",
            int,
            tgis_backend_priority=tgis_backend_priority,
            allow_none=True,
        )

        # Extract the TGIS backend instance
        local_initializer = MODEL_MANAGER.get_initializer(local_initializer_name)
        backends = local_initializer.backends
        if tgis_backend_priority is not None:
            error.value_check(
                "<NLP87928813E>",
                0 <= tgis_backend_priority < len(backends),
                "Invalid {}: {}",
                self._TGIS_BACKEND_PRIORITY_KEY,
                tgis_backend_priority,
            )
            self._tgis_backend = backends[tgis_backend_priority]
            error.value_check(
                "<NLP77150201E>",
                self._tgis_backend.backend_type == TGISBackend.backend_type,
                "Index {} is not a TGIS backend",
                tgis_backend_priority,
            )
        else:
            tgis_backend = None
            for backend in backends:
                if backend.backend_type == TGISBackend.backend_type:
                    tgis_backend = backend
                    break
            error.value_check(
                "<NLP96294266E>",
                tgis_backend is not None,
                "No TGIS backend found!",
            )
            self._tgis_backend = tgis_backend

    def find_model(
        self,
        model_path: str,
        **kwargs,
    ) -> Optional[ModuleConfig]:
        """Find the model if"""

        # Get a connection to this model in tgis
        log.debug2("Attempting to setup TGIS client for %s", model_path)
        if self._tgis_backend.get_connection(model_id=model_path) is None:
            log.debug2("TGIS cannot connect to model %s", model_path)
            return None

        # If connection is ok, set up the module config to point to the remote
        # TGIS text generation module
        cfg = ModuleConfig(
            {
                "module_id": TextGenerationTGIS.MODULE_ID,
                "module_class": TextGenerationTGIS.MODULE_CLASS,
                "name": TextGenerationTGIS.MODULE_NAME,
                "version": TextGenerationTGIS.MODULE_VERSION,
                "model_name": model_path,
            }
        )
        # Set a special indicator in the module config to use the backend that
        # this finder found. This will override the backend found by the local
        # initializer.
        cfg.tgis_backend = self._tgis_backend
        return cfg


model_finder_factory.register(TGISAutoFinder)



"""Tests for text-generation module
"""
# Standard
import os
import platform
import tempfile

# Third Party
import pytest
import torch

# First Party
from caikit.interfaces.nlp.data_model import GeneratedTextResult
import caikit

# Local
from caikit_nlp.data_model import GenerationTrainRecord
from caikit_nlp.modules.text_generation import TextGeneration
from caikit_nlp.resources.pretrained_model import HFAutoCausalLM, HFAutoSeq2SeqLM
from tests.fixtures import (
    CAUSAL_LM_MODEL,
    SEQ2SEQ_LM_MODEL,
    disable_wip,
    set_cpu_device,
)

### Stub Modules


def test_bootstrap_and_run_causallm():
    """Check if we can bootstrap and run causallm models"""

    model = TextGeneration.bootstrap(CAUSAL_LM_MODEL)

    sample_text = "Hello stub"
    generated_text = model.run(sample_text)
    assert isinstance(generated_text, GeneratedTextResult)


def test_bootstrap_and_run_seq2seq():
    """Check if we can bootstrap and run seq2seq models"""

    model = TextGeneration.bootstrap(SEQ2SEQ_LM_MODEL)

    sample_text = "Hello stub"
    generated_text = model.run(sample_text)
    assert isinstance(generated_text, GeneratedTextResult)


def test_bootstrap_and_save_model():
    """Check if we can bootstrap and save the model successfully"""

    model = TextGeneration.bootstrap(SEQ2SEQ_LM_MODEL)

    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        assert os.path.isfile(os.path.join(model_dir, "config.yml"))


def test_save_model_can_run():
    """Check if the model we bootstrap and save is able to load and run successfully"""
    model = TextGeneration.bootstrap(SEQ2SEQ_LM_MODEL)

    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        del model
        new_model = TextGeneration.load(model_dir)
        sample_text = "Hello stub"
        generated_text = new_model.run(sample_text)
        assert isinstance(generated_text, GeneratedTextResult)


############################## Training ################################


@pytest.mark.skipif(platform.processor() == "arm", reason="ARM training not supported")
def test_train_model_seq2seq(disable_wip, set_cpu_device):
    """Ensure that we can finetune a seq2seq model on some toy data for 1+
    steps & run inference."""
    train_kwargs = {
        "base_model": HFAutoSeq2SeqLM.bootstrap(
            model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
        ),
        "num_epochs": 1,
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
                GenerationTrainRecord(
                    input="@bar this is the worst idea ever.", output="complaint"
                ),
            ]
        ),
        "torch_dtype": torch.float32,
    }
    model = TextGeneration.train(**train_kwargs)
    assert isinstance(model.model, HFAutoSeq2SeqLM)

    # Ensure that we can get something out of it
    pred = model.run("@bar what a cute cat!")
    assert isinstance(pred, GeneratedTextResult)


@pytest.mark.skipif(platform.processor() == "arm", reason="ARM training not supported")
def test_train_model_save_and_load(disable_wip, set_cpu_device):
    """Ensure that we are able to save and load a finetuned model and execute inference on it"""
    train_kwargs = {
        "base_model": HFAutoSeq2SeqLM.bootstrap(
            model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
        ),
        "num_epochs": 1,
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                )
            ]
        ),
        "torch_dtype": torch.float32,
    }
    model = TextGeneration.train(**train_kwargs)
    assert isinstance(model.model, HFAutoSeq2SeqLM)
    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        new_model = TextGeneration.load(model_dir)
        sample_text = "Hello stub"
        generated_text = new_model.run(sample_text)
        assert isinstance(generated_text, GeneratedTextResult)


@pytest.mark.skipif(platform.processor() == "arm", reason="ARM training not supported")
def test_train_model_causallm(disable_wip, set_cpu_device):
    """Ensure that we can finetune a causal-lm model on some toy data for 1+
    steps & run inference."""
    train_kwargs = {
        "base_model": HFAutoCausalLM.bootstrap(
            model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
        ),
        "num_epochs": 1,
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
            ]
        ),
        "torch_dtype": torch.float32,
    }
    model = TextGeneration.train(**train_kwargs)
    assert isinstance(model.model, HFAutoCausalLM)

    # Ensure that we can get something out of it
    pred = model.run("@bar what a cute cat!")
    assert isinstance(pred, GeneratedTextResult)


############################## Inferencing flags ################################


@pytest.mark.skipif(platform.processor() == "arm", reason="ARM training not supported")
def test_train_model_causallm(disable_wip, set_cpu_device):
    """Ensure that we can finetune a causal-lm model on some toy data for 1+
    steps & run inference."""
    train_kwargs = {
        "base_model": HFAutoCausalLM.bootstrap(
            model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
        ),
        "num_epochs": 1,
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
            ]
        ),
        "torch_dtype": torch.float32,
    }
    model = TextGeneration.train(**train_kwargs)
    assert isinstance(model.model, HFAutoCausalLM)

    # Ensure that preserve_input_text returns input in output
    pred = model.run("@bar what a cute cat!", preserve_input_text=True)
    assert "@bar what a cute cat!" in pred.generated_text

    # Ensure that preserve_input_text set to False, removes input from output
    pred = model.run("@bar what a cute cat!", preserve_input_text=False)
    assert "@bar what a cute cat!" not in pred.generated_text


############################## Error Cases ################################


def test_zero_epoch_case(disable_wip):
    """Test to ensure 0 epoch training request doesn't explode"""
    train_kwargs = {
        "base_model": HFAutoSeq2SeqLM.bootstrap(
            model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
        ),
        "num_epochs": 0,
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
            ]
        ),
        "torch_dtype": torch.float32,
    }
    model = TextGeneration.train(**train_kwargs)
    assert isinstance(model.model, HFAutoSeq2SeqLM)


def test_run_tokenizer_not_implemented():
    with pytest.raises(NotImplementedError):
        model = TextGeneration.bootstrap(SEQ2SEQ_LM_MODEL)
        model.run_tokenizer("This text doesn't matter")



"""Tests for all pretrained model resource types; since most of the functionality for
these classes live through the base class, we test them all in this file.

Importantly, these tests do NOT depend on access to external internet to run correctly.

NOTE: If the subclasses become sufficiently complex, this test file should be split up.
"""
# Standard
from unittest.mock import patch

# Third Party
from datasets import IterableDataset as TransformersIterableDataset
from torch.utils.data import DataLoader
import pytest
import torch
import transformers

# First Party
from caikit.core.data_model import DataStream
import aconfig
import caikit

# Local
from caikit_nlp.data_model import GenerationTrainRecord
from caikit_nlp.resources.pretrained_model import HFAutoCausalLM, HFAutoSeq2SeqLM
from tests.fixtures import (
    CAUSAL_LM_MODEL,
    SEQ2SEQ_LM_MODEL,
    models_cache_dir,
    temp_cache_dir,
)


def test_boostrap_causal_lm(models_cache_dir):
    """Ensure that we can bootstrap a causal LM if we have download access."""
    # If we have an empty cachedir & do allow downloads, we should be able to init happily
    base_model = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
    )
    assert isinstance(base_model, HFAutoCausalLM)
    assert base_model.MODEL_TYPE is transformers.AutoModelForCausalLM
    assert base_model.TASK_TYPE == "CAUSAL_LM"


def test_boostrap_causal_lm_override_dtype(models_cache_dir):
    """Ensure that we can override the data type to init the model with at bootstrap time."""
    # If we have an empty cachedir & do allow downloads, we should be able to init happily,
    # and if we provide a data type override, that override should be applied at bootstrap time.
    base_model = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL,
        tokenizer_name=CAUSAL_LM_MODEL,
        torch_dtype=torch.float64,
    )
    assert isinstance(base_model, HFAutoCausalLM)
    assert base_model.MODEL_TYPE is transformers.AutoModelForCausalLM
    assert base_model.TASK_TYPE == "CAUSAL_LM"
    assert base_model.model.dtype is torch.float64


@patch("transformers.models.auto.tokenization_auto.AutoTokenizer.from_pretrained")
def test_boostrap_causal_lm_download_disabled(mock_tok_from_pretrained, temp_cache_dir):
    """Ensure that we can't try to download if downloads are disabled"""
    # NOTE: allow_downloads is false by default
    mock_tok_from_pretrained.side_effect = RuntimeError("It's a mock")
    with pytest.raises(RuntimeError):
        HFAutoCausalLM.bootstrap(model_name="foo/bar/baz", tokenizer_name="foo/bar/baz")
    kwargs = mock_tok_from_pretrained.call_args[1]
    assert kwargs["local_files_only"] == True


@patch("transformers.models.auto.tokenization_auto.AutoTokenizer.from_pretrained")
def test_boostrap_causal_lm_download_enabled(mock_tok_from_pretrained, temp_cache_dir):
    """Ensure that we can try to download if downloads are enabled."""
    with patch(
        "caikit_nlp.resources.pretrained_model.base.get_config",
        return_value=aconfig.Config({"allow_downloads": True}),
    ):
        mock_tok_from_pretrained.side_effect = RuntimeError("It's a mock")
        with pytest.raises(RuntimeError):
            HFAutoCausalLM.bootstrap(
                model_name="foo/bar/baz", tokenizer_name="foo/bar/baz"
            )
        kwargs = mock_tok_from_pretrained.call_args[1]
        assert kwargs["local_files_only"] == False


def test_boostrap_model_path(models_cache_dir):
    """Ensure that we can bootstrap works with loading a model from local directory"""
    # If we have an empty cachedir & do allow downloads, we should be able to init happily
    base_model = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL,
    )
    assert isinstance(base_model, HFAutoCausalLM)
    assert base_model.MODEL_TYPE is transformers.AutoModelForCausalLM
    assert base_model.TASK_TYPE == "CAUSAL_LM"


### Tests for tokenization behaviors
SAMPLE_TRAINING_DATA = caikit.core.data_model.DataStream.from_iterable(
    [
        GenerationTrainRecord(input="This is a sequence", output="This is a target"),
        GenerationTrainRecord(input="I am just a little", output="bit tired"),
    ]
)

# Causal LM tokenization strategies
### 1. Tests for Causal LM tokenization chunking
def test_causal_lm_tokenize_func_contains_wrapped_stream(models_cache_dir):
    """Ensure the Causal LM tokenize func produces a wrapped stream that can be flattened."""
    causal_lm = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
    )
    (tok_func, requires_unwrapping) = causal_lm.build_task_tokenize_closure(
        tokenizer=causal_lm.tokenizer,
        max_source_length=100,
        max_target_length=100,
        verbalizer="{{input}}",
        use_seq2seq_approach=False,
    )
    map_stream = SAMPLE_TRAINING_DATA.map(tok_func)
    # Since tok_func for causal lm creates a datastream, we should get a stream
    # back; make sure that it's a stream of streams before proceeding
    assert requires_unwrapping is True
    assert isinstance(map_stream, caikit.core.data_model.DataStream)
    assert isinstance(map_stream.peek(), caikit.core.data_model.DataStream)
    # Ensure that we can unwrap the stream
    unwrapped_stream = map_stream.flatten()
    assert isinstance(
        unwrapped_stream.peek(), transformers.tokenization_utils_base.BatchEncoding
    )


# Key cases here are:
# 1 - simplest and minimal case
# 3 - because the concat sequence is length 17, so we have a remainder
# 100 - which is much larger than the concatenated seq and should yield one chunk
@pytest.mark.parametrize(
    "chunk_size,drop_remainder",
    [(1, True), (1, False), (3, True), (3, False), (100, True), (100, False)],
)
def test_causal_lm_tok_output_correctness(models_cache_dir, chunk_size, drop_remainder):
    """Validate the tokenized results for the chunked language modeling objective."""
    causal_lm = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
    )
    sample = GenerationTrainRecord(
        input="Hello world", output="How are you doing today?!"
    )
    (tok_func, _) = causal_lm.build_task_tokenize_closure(
        tokenizer=causal_lm.tokenizer,
        max_source_length=100,
        max_target_length=100,
        verbalizer="{{input}}",
        task_ids=0,
        use_seq2seq_approach=False,
        chunk_size=chunk_size,
        drop_remainder=drop_remainder,
    )
    input_tok = causal_lm.tokenizer.encode(sample.input)
    output_tok = causal_lm.tokenizer.encode(sample.output)
    concat_tok = input_tok + output_tok
    tok_stream = tok_func(sample)
    # Ensure we get one token per output in our stream
    assert isinstance(tok_stream, caikit.core.data_model.DataStream)
    # Figure out how many chunks we should have, including if we have a remainder
    has_remainder = False
    if len(concat_tok) > chunk_size:
        num_expected_chunks = len(concat_tok) // chunk_size
        # Should only care about the remainder if we are not dropping it
        if num_expected_chunks * chunk_size != len(concat_tok) and not drop_remainder:
            has_remainder = True
    else:
        num_expected_chunks = 1
        chunk_size = len(concat_tok)
    tok_list = list(tok_stream)
    assert len(tok_list) == num_expected_chunks + has_remainder
    # Check all full chunks. Note that we always attend to everything
    for idx in range(num_expected_chunks):
        assert len(tok_list[idx]["attention_mask"]) == chunk_size
        assert len(tok_list[idx]["input_ids"]) == chunk_size
        assert all(atn == 1 for atn in tok_list[idx]["attention_mask"])
        assert tok_list[idx]["task_ids"] == 0
    # Check the remainder; lists should be the same length, but less than the chunk size
    if has_remainder:
        remainder = tok_list[-1]
        assert len(remainder["attention_mask"]) == len(remainder["input_ids"])
        assert len(remainder["input_ids"]) < chunk_size
        assert all(atn == 1 for atn in remainder["attention_mask"])


def test_causal_lm_batch_tokenization(models_cache_dir):
    """Ensure that we can batch process causal lm inputs correctly."""
    causal_lm = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
    )
    train_stream = DataStream.from_iterable(
        [
            GenerationTrainRecord(input="hello there", output="world"),
            GenerationTrainRecord(input="how", output="today"),
        ]
    )
    fn_kwargs = {
        "tokenizer": causal_lm.tokenizer,
        "max_source_length": 10,
        "max_target_length": 10,
        "use_seq2seq_approach": False,
    }
    # Create an iterable dataset by batching...
    def get(train_stream):
        for data in train_stream:
            yield {"input": data.input, "output": data.output}

    dataset = TransformersIterableDataset.from_generator(
        get, gen_kwargs={"train_stream": train_stream}
    )
    batched_dataset = dataset.map(
        causal_lm.tokenize_function,
        fn_kwargs=fn_kwargs,
        batched=True,
        remove_columns=["input", "output"],
    )

    # Do the same thing with no batching via tokenize closure + unwrapping
    tok_func = causal_lm.build_task_tokenize_closure(**fn_kwargs)[0]
    mapped_indiv_stream = train_stream.map(tok_func).flatten()
    for indiv_res, batched_res in zip(mapped_indiv_stream, batched_dataset):
        # All keys should match (input ids, attention mask)
        assert indiv_res.keys() == batched_res.keys()
        # And all of their values should be the same
        for k in indiv_res:
            assert indiv_res[k] == batched_res[k]


### 2. Tests for causal LM framed as a seq2seq problem
# NOTE: For these tests, we should be careful to always test left and right padding
@pytest.mark.parametrize(
    "padding_side",
    ["left", "right"],
)
def test_causal_lm_as_a_sequence_problem_no_truncation(models_cache_dir, padding_side):
    causal_lm = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
    )
    sample = GenerationTrainRecord(
        input="Hello world", output="How are you doing today?!"
    )
    max_lengths = 20
    # First, build the output we expect for left / right respectively...
    input_tok = causal_lm.tokenizer.encode(sample.input)
    output_tok = causal_lm.tokenizer.encode(sample.output) + [
        causal_lm.tokenizer.eos_token_id
    ]
    concat_res = input_tok + output_tok
    masked_res = ([-100] * len(input_tok)) + output_tok

    # This must true because otherwise no padding was needed, e.g., truncation
    assert len(input_tok) < max_lengths
    assert len(output_tok) < (max_lengths + 1)
    pads_needed = (1 + 2 * max_lengths) - len(concat_res)
    if causal_lm.tokenizer.padding_side.lower() == "left":
        expected_input_ids = torch.tensor(
            [causal_lm.tokenizer.pad_token_id] * pads_needed + concat_res
        )
        expected_attn_mask = torch.tensor([0] * pads_needed + [1] * len(concat_res))
        expected_labels = torch.tensor([-100] * pads_needed + masked_res)
    else:
        expected_input_ids = torch.tensor(
            concat_res + [causal_lm.tokenizer.pad_token_id] * pads_needed
        )
        expected_attn_mask = torch.tensor([1] * len(concat_res) + [0] * pads_needed)
        expected_labels = torch.tensor(masked_res + [-100] * pads_needed)

    # Now build the analogous tokenizer closure and compare the tensors
    (tok_func, _) = causal_lm.build_task_tokenize_closure(
        tokenizer=causal_lm.tokenizer,
        max_source_length=max_lengths,
        max_target_length=max_lengths,
        verbalizer="{{input}}",
        task_ids=0,
        use_seq2seq_approach=True,
    )
    tok_res = tok_func(sample)
    assert tok_res["task_ids"] == 0
    assert torch.all(tok_res["input_ids"] == expected_input_ids)
    assert torch.all(tok_res["attention_mask"] == expected_attn_mask)
    assert torch.all(tok_res["labels"] == expected_labels)


### Tests for Seq2Seq tokenization
def test_seq2seq_tokenize_func_contains_unwrapped_stream(models_cache_dir):
    """Ensure the seq2seq tokenizer produces an unwrapped stream; not flattening needed."""
    seq2seq = HFAutoSeq2SeqLM.bootstrap(
        model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
    )
    (tok_func, requires_unwrapping) = seq2seq.build_task_tokenize_closure(
        tokenizer=seq2seq.tokenizer,
        max_source_length=100,
        max_target_length=100,
        verbalizer="{{input}}",
        task_ids=0,
    )
    map_stream = SAMPLE_TRAINING_DATA.map(tok_func)
    # Since we don't require unwrapping, i.e., each input sample just produces 1,
    # result, we should just get a stream of batch encodings we can use directly.
    assert requires_unwrapping is False
    assert isinstance(map_stream, caikit.core.data_model.DataStream)
    assert isinstance(
        map_stream.peek(), transformers.tokenization_utils_base.BatchEncoding
    )


def test_seq2seq_tok_output_correctness(models_cache_dir):
    """Validate the correctness of the attention mask for the seq2seq task."""
    seq2seq = HFAutoSeq2SeqLM.bootstrap(
        model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
    )
    sample = GenerationTrainRecord(
        input="This len does not matter", output="and this one doesn't either!"
    )
    (tok_func, _) = seq2seq.build_task_tokenize_closure(
        tokenizer=seq2seq.tokenizer,
        max_source_length=20,
        max_target_length=20,
        verbalizer="{{input}}",
        task_ids=0,
    )
    input_tok = seq2seq.tokenizer.encode(sample.input)

    tok_sample = tok_func(sample)
    # Ensure we get one seq2seq; i.e., the result should NOT be a stream,
    # and we should only be attending to the tokens from the input sequence.
    assert isinstance(tok_sample, transformers.tokenization_utils_base.BatchEncoding)
    assert sum(tok_sample["attention_mask"]) == len(input_tok)
    # Ensure we support MPT
    assert hasattr(tok_sample, "task_ids")
    assert tok_sample["task_ids"] == 0


### Tests for collator compatability
# These tests should validate that we can use our tokenization function to
# build torch loaders around datasets using different collators.
# TODO: Expand to cover transformer datasets, i.e., what is produced by
# text gen preprocessing functions. For now, they only check the minimal
# case with the default data collator.
@pytest.mark.parametrize(
    "collator_fn",
    [transformers.default_data_collator],
)
def test_loader_can_batch_list_of_seq2seq_outputs(collator_fn):
    # Build the dataset
    train_stream = DataStream.from_iterable(
        [
            GenerationTrainRecord(input="hello world", output="how are you today?"),
            GenerationTrainRecord(input="goodbye", output="world"),
            GenerationTrainRecord(input="good morning", output="have a good day"),
            GenerationTrainRecord(input="good night", output="have nice dreams"),
        ]
    )
    seq2seq = HFAutoSeq2SeqLM.bootstrap(
        model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
    )
    (tok_func, _) = seq2seq.build_task_tokenize_closure(
        tokenizer=seq2seq.tokenizer,
        max_source_length=20,
        max_target_length=20,
        verbalizer="{{input}}",
        task_ids=0,
    )
    tok_results = [tok_func(x) for x in list(train_stream)]
    dl = DataLoader(
        tok_results,
        shuffle=False,
        batch_size=2,
        collate_fn=collator_fn,
    )
    # Loader should create 2 batches
    loader_list = list(dl)
    assert len(loader_list) == 2


@pytest.mark.parametrize(
    "collator_fn",
    [transformers.default_data_collator],
)
def test_loader_can_batch_list_of_causal_lm_outputs(collator_fn):
    # Build the dataset
    train_stream = DataStream.from_iterable(
        [
            GenerationTrainRecord(input="hello world", output="how are you today?"),
            GenerationTrainRecord(input="goodbye", output="world"),
            GenerationTrainRecord(input="good morning", output="have a good day"),
            GenerationTrainRecord(input="good night", output="have nice dreams"),
        ]
    )
    causal_lm = HFAutoCausalLM.bootstrap(
        model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
    )
    (tok_func, _) = causal_lm.build_task_tokenize_closure(
        tokenizer=causal_lm.tokenizer,
        max_source_length=20,
        max_target_length=20,
        verbalizer="{{input}}",
        task_ids=0,
        use_seq2seq_approach=True,
    )
    tok_results = [tok_func(x) for x in list(train_stream)]
    dl = DataLoader(
        tok_results,
        shuffle=False,
        batch_size=2,
        collate_fn=collator_fn,
    )
    # Loader should create 2 batches
    loader_list = list(dl)
    assert len(loader_list) == 2



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from collections.abc import Sized
from enum import Enum, auto
from typing import Any, Callable, Dict, List, NamedTuple, Optional, TypeVar, Union
import importlib
import os
import time

# Third Party
from torch.backends import mps
from transformers import BatchEncoding
import numpy as np
import torch

# First Party
from caikit import get_config
from caikit.core import ModuleBase, ModuleConfig, ModuleSaver, module
from caikit.core.data_model.json_dict import JsonDict
from caikit.core.exceptions import error_handler
from caikit.interfaces.common.data_model.vectors import ListOfVector1D, Vector1D
from caikit.interfaces.nlp.data_model import (
    EmbeddingResult,
    EmbeddingResults,
    RerankResult,
    RerankResults,
    RerankScore,
    RerankScores,
    SentenceSimilarityResult,
    SentenceSimilarityResults,
    SentenceSimilarityScores,
)
from caikit.interfaces.nlp.tasks import (
    EmbeddingTask,
    EmbeddingTasks,
    RerankTask,
    RerankTasks,
    SentenceSimilarityTask,
    SentenceSimilarityTasks,
)
import alog

# Local
from caikit_nlp.modules.text_embedding.utils import env_val_to_bool, env_val_to_int

logger = alog.use_channel("TXT_EMB")
error = error_handler.get(logger)


# To avoid dependency problems, make sentence-transformers an optional import and
# defer any ModuleNotFoundError until someone actually tries to init a model with this module.
try:
    sentence_transformers = importlib.import_module("sentence_transformers")
    # Third Party
    from sentence_transformers import SentenceTransformer
    from sentence_transformers.util import batch_to_device, cos_sim, dot_score
    from sentence_transformers.util import (
        normalize_embeddings as normalize,  # avoid parameter shadowing
    )
    from sentence_transformers.util import semantic_search
except ModuleNotFoundError:
    # When it is not available, create a dummy that raises an error on attempted init()
    class SentenceTransformerNotAvailable:
        def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
            # Will reproduce the ModuleNotFoundError if/when anyone actually tries this module/model
            importlib.import_module("sentence_transformers")

    SentenceTransformer = SentenceTransformerNotAvailable

embedding_cfg = get_config().get("embedding", {})

AUTOCAST = env_val_to_bool(val=embedding_cfg.get("autocast"))
IPEX = env_val_to_bool(val=embedding_cfg.get("ipex"))
PT2_COMPILE = env_val_to_bool(val=embedding_cfg.get("pt2_compile"))
RETRIES = env_val_to_int(val=embedding_cfg.get("retries"), default=0)
BATCH_SIZE = env_val_to_int(val=embedding_cfg.get("batch_size"), default=0)
NO_IMPLICIT_TRUNCATION = env_val_to_bool(
    val=embedding_cfg.get("implicit_truncation_errors", True)
)
DEVICE = embedding_cfg.get("device", "")

RT = TypeVar("RT")  # return type


class EmbeddingResultTuple(NamedTuple):
    """Output of SentenceTransformerWithTruncate.encode()"""

    embedding: np.ndarray
    input_token_count: int


class TruncatedTokensTuple(NamedTuple):
    """Output of SentenceTransformerWithTruncate._truncate_input_tokens()"""

    tokenized: BatchEncoding
    input_token_count: int


@module(
    "eeb12558-b4fa-4f34-a9fd-3f5890e9cd3f",
    "EmbeddingModule",
    "0.0.1",
    tasks=[
        EmbeddingTask,
        EmbeddingTasks,
        SentenceSimilarityTask,
        SentenceSimilarityTasks,
        RerankTask,
        RerankTasks,
    ],
)
class EmbeddingModule(ModuleBase):
    # Retry count if enabled to try again (was for thread contention errors)
    RETRY_COUNT = max(RETRIES, 0)  # Ensure non-negative, before using in loop!

    _ARTIFACTS_PATH_KEY = "artifacts_path"
    _ARTIFACTS_PATH_DEFAULT = "artifacts"

    def __init__(
        self,
        model: SentenceTransformer,
    ):
        super().__init__()
        self.model = model

    @classmethod
    def load(cls, model_path: str, *args, **kwargs) -> "EmbeddingModule":
        """Load model

        Args:
            model_path: str
                Path to the config dir under the model_id (where the config.yml lives)

        Returns:
            EmbeddingModule
                Instance of this class built from the model.
        """

        config = ModuleConfig.load(model_path)
        artifacts_path = config.get(cls._ARTIFACTS_PATH_KEY)

        error.value_check(
            "<NLP07391618E>",
            artifacts_path,
            ValueError(f"Model config missing '{cls._ARTIFACTS_PATH_KEY}'"),
        )

        artifacts_path = os.path.abspath(os.path.join(model_path, artifacts_path))
        error.dir_check("<NLP34197772E>", artifacts_path)

        ipex = cls._get_ipex(IPEX)
        device = cls._select_device(ipex, DEVICE)
        model = SentenceTransformerWithTruncate(
            model_name_or_path=artifacts_path, device=device
        )
        model.eval()  # required for IPEX at least
        if device is not None:
            model.to(torch.device(device))
        model = EmbeddingModule._optimize(model, ipex, device, AUTOCAST, PT2_COMPILE)

        # Validate model with any encode test (simple and hardcoded for now).
        # This gets some of the first-time inference cost out of the way.
        # This avoids using the tokenizer (for truncation) before it is ready.
        model.encode("warmup")

        return cls(model)

    @property
    def public_model_info(cls) -> Dict[str, Any]:  # pylint: disable=no-self-argument
        """Helper property to return public metadata about a specific Model. This
        function is separate from `metdata` as that contains the entire ModelConfig
        which might not want to be shared/exposed.

        Returns:
            Dict[str, str]: A dictionary of this models's public metadata
        """
        return {
            "max_seq_length": cls.model.max_seq_length,
            "sentence_embedding_dimension": cls.model.get_sentence_embedding_dimension(),
        }

    @classmethod
    def _get_ipex(cls, ipex_flag):
        """Get IPEX optimization library if enabled and available, else return False

        Returns ipex library or False
        """
        ret = False

        # Enabled by environment variable
        # When IPEX is not false, attempt to import the library and use it.
        if ipex_flag:
            try:
                ret = importlib.import_module("intel_extension_for_pytorch")
            except Exception as ie:  # pylint: disable=broad-exception-caught
                # We don't require the module so catch, log, proceed to return False
                msg = (
                    f"IPEX enabled in env, but skipping ipex.optimize() because "
                    f"import intel_extension_for_pytorch failed with exception: {ie}"
                )
                logger.warning(msg, exc_info=True)

        return ret

    @staticmethod
    def _select_device(use_ipex, device):
        """Use environment variables and availability to determine the device to use"""
        if use_ipex:
            # If enabled, use "xpu" (IPEX on GPU instead of IPEX on CPU)
            if device == "xpu":
                return "xpu"
        elif device == "mps" and mps.is_built() and mps.is_available():
            # Never use on ipex, but otherwise use mps if enabled and available
            return "mps"

        return "cuda" if torch.cuda.is_available() else None

    @staticmethod
    def _get_backend(use_ipex, use_device):
        """Determine the backend to use for torch compile.

        Considers global ipex if enabled first, next mps device, finally defaults.

        Returns the backend for torch.compile()
        """
        if use_ipex:
            return "ipex"
        if use_device == "mps":
            return mps
        return "inductor"  # default backend

    @staticmethod
    def _optimize(model, ipex, device, autocast, pt2_compile):
        if ipex:
            if autocast:  # IPEX performs best with autocast using bfloat16
                model = ipex.optimize(
                    model, dtype=torch.bfloat16, weights_prepack=False
                )
            else:
                model = ipex.optimize(model, weights_prepack=False)

        # torch.compile won't work everywhere, but when set we'll try it
        if pt2_compile:
            backend = EmbeddingModule._get_backend(ipex, device)
            try:
                model = torch.compile(model, backend=backend, mode="max-autotune")
            except Exception as e:  # pylint: disable=broad-exception-caught
                # Not always supported (e.g. in a python version) so catch, log, proceed.
                warn_msg = (
                    f"PT2_COMPILE enabled, but continuing without torch.compile() "
                    f"because it failed with exception: {e}"
                )
                logger.warning(warn_msg, exc_info=True)
        return model

    def _with_retry(self, fn: Callable[..., RT], *args, **kwargs) -> RT:
        first_exception = None
        for count in range(1 + self.RETRY_COUNT):  # try once plus retries (if needed)
            try:
                return fn(*args, **kwargs)
            except Exception as e:  # pylint: disable=broad-exception-caught
                if first_exception is None:
                    first_exception = e
                if self.RETRY_COUNT > 0:
                    warn_msg = f"Try {count + 1}: {fn} failed due to: {e}"
                    logger.warning("<NLP54902271W>", warn_msg, exc_info=True)
                    if count + 1 < self.RETRY_COUNT:
                        time.sleep(0.1 * (count * 2))

        # If above return did not happen, raise the first exception
        error.log_raise(
            log_code="<NLP13096081E>",
            exception=first_exception,
        )

    def _encode_with_retry(
        self, *args, **kwargs
    ) -> Union[EmbeddingResultTuple, List[torch.Tensor], np.ndarray, torch.Tensor]:
        """All encode calls should use this for consistent param adding and retry loop"""

        # Add the batch_size kwarg if not passed in and given a usable BATCH_SIZE
        if BATCH_SIZE > 0:
            if kwargs is None:
                kwargs = {}
            if "batch_size" not in kwargs:
                kwargs["batch_size"] = BATCH_SIZE

        if isinstance(self.model, SentenceTransformerWithTruncate):
            kwargs[
                "implicit_truncation_errors"
            ] = NO_IMPLICIT_TRUNCATION  # config/env overrides default
            return self._with_retry(self.model.encode, *args, **kwargs)

        # Else...
        # It's possible to init with a model that doesn't have the added kwargs.
        # E.g. a SentenceTransformer or other transformer model. Remove those kwargs!
        # This is not the normal use case but at least don't pass invalid kwargs, to encode()
        # and don't return the unexpected tuple (adding token count).
        if "truncate_input_tokens" in kwargs:
            del kwargs["truncate_input_tokens"]
        if "return_token_count" in kwargs:
            del kwargs["return_token_count"]
        if "implicit_truncation_errors" in kwargs:
            del kwargs["implicit_truncation_errors"]
        return self._with_retry(self.model.encode, *args, **kwargs)

    @EmbeddingTask.taskmethod()
    def run_embedding(
        self,
        text: str,
        truncate_input_tokens: Optional[int] = 0,
    ) -> EmbeddingResult:
        """Get embedding for a string.
        Args:
            text: str
                Input text to be processed
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this is disabled (returns texts without processing).
                If zero or greater than the model's maximum, then this is a test
                to see if truncation is needed. If needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the tokens and
                decode them to return truncated strings that can be used with this model.
        Returns:
            EmbeddingResult: the result vector nicely wrapped up
        """
        error.type_check("<NLP27491611E>", str, text=text)

        embeddings, input_token_count = self._encode_with_retry(
            text,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
        )
        return EmbeddingResult(
            result=Vector1D.from_vector(embeddings),
            producer_id=self.PRODUCER_ID,
            input_token_count=input_token_count,
        )

    @EmbeddingTasks.taskmethod()
    def run_embeddings(
        self,
        texts: List[str],
        truncate_input_tokens: Optional[int] = 0,
    ) -> EmbeddingResults:
        """Get embedding vectors for texts.
        Args:
            texts: List[str]
                List of input texts to be processed
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this is disabled (returns texts without processing).
                If zero or greater than the model's maximum, then this is a test
                to see if truncation is needed. If needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the tokens and then
                decode them to return truncated strings that can be used with this model.
        Returns:
            EmbeddingResults: List of vectors. One for each input text (in order).
             Each vector is a list of floats (supports various float types).
        """
        if isinstance(
            texts, str
        ):  # encode allows str, but the result would lack a dimension
            texts = [texts]

        embeddings, input_token_count = self._encode_with_retry(
            texts,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
        )
        vectors = [Vector1D.from_vector(e) for e in embeddings]

        return EmbeddingResults(
            results=ListOfVector1D(vectors=vectors),
            producer_id=self.PRODUCER_ID,
            input_token_count=input_token_count,
        )

    @SentenceSimilarityTask.taskmethod()
    def run_sentence_similarity(
        self,
        source_sentence: str,
        sentences: List[str],
        truncate_input_tokens: Optional[int] = 0,
    ) -> SentenceSimilarityResult:
        """Get similarity scores for each of sentences compared to the source_sentence.
        Args:
            source_sentence: str
            sentences: List[str]
                Sentences to compare to source_sentence
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this is disabled (returns texts without processing).
                If zero or greater than the model's maximum, then this is a test
                to see if truncation is needed. If needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the tokens and then
                decode them to return truncated strings that can be used with this model.
        Returns:
            SentenceSimilarityResult: Similarity scores for each sentence.
        """

        source_embedding, source_token_count = self._encode_with_retry(
            source_sentence,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
        )
        embeddings, sentences_token_count = self._encode_with_retry(
            sentences,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
        )

        input_token_count = source_token_count + sentences_token_count
        res = cos_sim(source_embedding, embeddings)

        return SentenceSimilarityResult(
            result=SentenceSimilarityScores(scores=res.tolist()[0]),
            producer_id=self.PRODUCER_ID,
            input_token_count=input_token_count,
        )

    @SentenceSimilarityTasks.taskmethod()
    def run_sentence_similarities(
        self,
        source_sentences: List[str],
        sentences: List[str],
        truncate_input_tokens: Optional[int] = 0,
    ) -> SentenceSimilarityResults:
        """Run sentence-similarities on model.
        Args:
            source_sentences: List[str]
            sentences: List[str]
                Sentences to compare to source_sentences
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this is disabled (returns texts without processing).
                If zero or greater than the model's maximum, then this is a test
                to see if truncation is needed. If needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the tokens and then
                decode them to return truncated strings that can be used with this model.
        Returns:
            SentenceSimilarityResults: Similarity scores for each source sentence in order.
                Each one contains the source-sentence's score for each sentence in order.
        """

        source_embedding, source_token_count = self._encode_with_retry(
            source_sentences,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
        )
        embeddings, sentences_token_count = self._encode_with_retry(
            sentences,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
        )

        input_token_count = source_token_count + sentences_token_count
        res = cos_sim(source_embedding, embeddings)
        float_list_list = res.tolist()

        return SentenceSimilarityResults(
            results=[SentenceSimilarityScores(fl) for fl in float_list_list],
            producer_id=self.PRODUCER_ID,
            input_token_count=input_token_count,
        )

    @RerankTask.taskmethod()
    def run_rerank_query(
        self,
        query: str,
        documents: List[JsonDict],
        top_n: Optional[int] = None,
        truncate_input_tokens: Optional[int] = 0,
        return_documents: bool = True,
        return_query: bool = True,
        return_text: bool = True,
    ) -> RerankResult:
        """Rerank the documents returning the most relevant top_n in order for this query.
        Args:
            query: str
                Query is the source string to be compared to the text of the documents.
            documents:  List[JsonDict]
                Each document is a dict. The text value is used for comparison to the query.
                If there is no text key, then _text is used and finally default is "".
            top_n:  Optional[int]
                Results for the top n most relevant documents will be returned.
                If top_n is not provided or (not > 0), then all are returned.
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this is disabled (returns texts without processing).
                If zero or greater than the model's maximum, then this is a test
                to see if truncation is needed. If needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the tokens and then
                decode them to return truncated strings that can be used with this model.
            return_documents:  bool
                Default True
                Setting to False will disable returning of the input document (index is returned).
            return_query:  bool
                Default True
                Setting to False will disable returning of the query (results are in query order)
            return_text:  bool
                Default True
                Setting to False will disable returning of document text string that was used.
        Returns:
            RerankResult
                Returns the (top_n) scores in relevance order (most relevant first).
                The results always include a score and index which may be used to find the document
                in the original documents list. Optionally, the results also contain the entire
                document with its score (for use in chaining) and for convenience the query and
                text used for comparison may be returned.

        """

        error.type_check(
            "<NLP05323654E>",
            str,
            query=query,
        )

        results = self.run_rerank_queries(
            queries=[query],
            documents=documents,
            top_n=top_n,
            truncate_input_tokens=truncate_input_tokens,
            return_documents=return_documents,
            return_queries=return_query,
            return_text=return_text,
        )

        if results.results:
            return RerankResult(
                result=results.results[0],
                producer_id=self.PRODUCER_ID,
                input_token_count=results.input_token_count,
            )

        RerankResult(
            result=RerankScore(
                scores=[],
                query=query if return_query else None,
            ),
            producer_id=self.PRODUCER_ID,
            input_token_count=results.input_token_count,
        )

    @RerankTasks.taskmethod()
    def run_rerank_queries(
        self,
        queries: List[str],
        documents: List[JsonDict],
        top_n: Optional[int] = None,
        truncate_input_tokens: Optional[int] = 0,
        return_documents: bool = True,
        return_queries: bool = True,
        return_text: bool = True,
    ) -> RerankResults:
        """Rerank the documents returning the most relevant top_n in order for each of the queries.
        Args:
            queries: List[str]
                Each of the queries will be compared to the text of each of the documents.
            documents:  List[JsonDict]
                Each document is a dict. The text value is used for comparison to the query.
                If there is no text key, then _text is used and finally default is "".
            top_n:  Optional[int]
                Results for the top n most relevant documents will be returned.
                If top_n is not provided or (not > 0), then all are returned.
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this is disabled (returns texts without processing).
                If zero or greater than the model's maximum, then this is a test
                to see if truncation is needed. If needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the tokens and then
                decode them to return truncated strings that can be used with this model.
            return_documents:  bool
                Default True
                Setting to False will disable returning of the input document (index is returned).
            return_queries:  bool
                Default True
                Setting to False will disable returning of the query (results are in query order)
            return_text:  bool
                Default True
                Setting to False will disable returning of document text string that was used.
        Returns:
            RerankResults
                For each query in queries (in the original order)...
                Returns the (top_n) scores in relevance order (most relevant first).
                The results always include a score and index which may be used to find the document
                in the original documents list. Optionally, the results also contain the entire
                document with its score (for use in chaining) and for convenience the query and
                text used for comparison may be returned.
        """

        error.type_check(
            "<NLP09038249E>",
            list,
            queries=queries,
            documents=documents,
        )

        error.value_check(
            "<NLP24788937E>",
            queries and documents,
            "Cannot rerank without a query and at least one document",
        )

        if top_n is None or top_n < 1:
            top_n = len(documents)

        # Using input document dicts so get "text" else "_text" else default to ""
        def get_text(doc):
            return doc.get("text") or doc.get("_text", "")

        doc_texts = [get_text(doc) for doc in documents]

        doc_embeddings, doc_token_count = self._encode_with_retry(
            doc_texts,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
            convert_to_tensor=True,
        )
        doc_embeddings = normalize(doc_embeddings.to(self.model.device))

        query_embeddings, query_token_count = self._encode_with_retry(
            queries,
            truncate_input_tokens=truncate_input_tokens,
            return_token_count=True,
            convert_to_tensor=True,
        )
        query_embeddings = normalize(query_embeddings.to(self.model.device))

        res = semantic_search(
            query_embeddings, doc_embeddings, top_k=top_n, score_function=dot_score
        )

        # Fixup result dicts
        for r in res:
            for x in r:
                # Renaming corpus_id to index
                corpus_id = x.pop("corpus_id")
                x["index"] = corpus_id
                # Optionally adding the original document and/or just the text that was used
                if return_documents:
                    x["document"] = documents[corpus_id]
                if return_text:
                    x["text"] = get_text(documents[corpus_id])

        def add_query(q):
            return queries[q] if return_queries else None

        results = [
            RerankScores(
                query=add_query(q),
                scores=[RerankScore(**x) for x in r],
            )
            for q, r in enumerate(res)
        ]
        input_token_count = doc_token_count + query_token_count

        return RerankResults(
            results=results,
            producer_id=self.PRODUCER_ID,
            input_token_count=input_token_count,
        )

    @classmethod
    def bootstrap(cls, model_name_or_path: str) -> "EmbeddingModule":
        """Bootstrap a sentence-transformers model

        Args:
            model_name_or_path: str
                Model name (Hugging Face hub) or path to model to load.
        """
        return cls(model=SentenceTransformer(model_name_or_path=model_name_or_path))

    def save(self, model_path: str, *args, **kwargs):
        """Save model using config in model_path

        Args:
            model_path: str
                Path to model config
        """

        error.type_check("<NLP82314992E>", str, model_path=model_path)
        model_config_path = model_path.strip()
        error.value_check(
            "<NLP40145207E>",
            model_config_path,
            f"model_path '{model_config_path}' is invalid",
        )

        model_config_path = os.path.abspath(
            model_config_path.strip()
        )  # No leading/trailing spaces sneaky weirdness

        # Only allow new dirs because there are not enough controls to safely update in-place
        os.makedirs(model_config_path, exist_ok=False)

        saver = ModuleSaver(
            module=self,
            model_path=model_config_path,
        )
        artifacts_path = self._ARTIFACTS_PATH_DEFAULT
        saver.update_config({self._ARTIFACTS_PATH_KEY: artifacts_path})

        # Save the model
        self.model.save(os.path.join(model_config_path, artifacts_path))

        # Save the config
        ModuleConfig(saver.config).save(model_config_path)


def get_sample_start_indexes(tokenized: BatchEncoding) -> List[int]:
    """Returns a list containing the index for the first encoding of each sample
    contained in tokenized."""

    # When truncating occurs a sample is split across multiple encodings
    # ie. len(tokenized.encodings) > the number of text samples input for tokenization

    # Knowing the encoding index of where each sample's first encoding is located allows us to
    # access the encodings for individual samples

    # note: tokenized["overflow_to_sample_mapping"] is a torch.Tensor

    samples_start_indexes: Dict[int, int] = {}
    for i, tensor_sample in enumerate(tokenized["overflow_to_sample_mapping"]):
        int_sample = int(tensor_sample)
        if int_sample not in samples_start_indexes:
            samples_start_indexes[int_sample] = i

    return list(samples_start_indexes.values())


class TruncateCountBehavior(Enum):
    ONLY = auto()
    ALL = auto()
    IGNORE = auto()


def sum_token_count(
    tokenized: BatchEncoding,
    truncate_only: bool,
) -> int:
    """Returns the number of non-special tokens.
    Args:
        tokenized: BatchEncoding
        truncate_only: bool
    Returns:
        Int total of all tokens contained in tokenized.
    """
    # Encoding objects have various attributes of note:
    # - tokens: list of tokens (sub-parts of the input strings after word/subword
    #       splitting and before conversion to integer indices)
    # - attention_mask: List of indices specifying which tokens should be attended to
    #       by the model. Note that [PAD] = 0, while [CLS] / [SEP] = 1
    # - special_tokens_mask: List of 0s and 1s, with 1 specifying added special tokens
    #       and 0 specifying regular sequence tokens

    error.type_check(
        "<NLP82314993E>",
        BatchEncoding,
        tokenized=tokenized,
    )
    error.value_check(
        "<NLP82314995E>",
        tokenized.encodings,
        "Number of tokenized encodings is only known when a non-python tokenizer is used",
    )

    token_count = 0

    if truncate_only:
        # Only sum the length for the 1st encoding of each sample
        samples_start_idx = get_sample_start_indexes(tokenized)

        token_count = sum(
            (
                x
                for idx in samples_start_idx
                for x in tokenized.encodings[idx].attention_mask
            )
        )
    else:
        # Sum the length of all encodings for all samples
        for encoding in tokenized.encodings:
            token_count += sum(encoding.attention_mask)

    return token_count


class SentenceTransformerWithTruncate(SentenceTransformer):
    def _truncate_input_tokens(
        self,
        truncate_input_tokens: int,
        texts: List[str],
        implicit_truncation_errors: bool = True,
    ) -> TruncatedTokensTuple:
        """Truncate input tokens
        Args:
            truncate_input_tokens: int
                Truncation length for input tokens.
                If less than zero, this truncation is left up to the tokenizer default (model max).
                If zero or greater than the model's maximum, then this is used as a test
                to see if truncation is needed. If needed is needed, an exception is thrown.
                Otherwise, we take this usable truncation limit to truncate the input tokens.
            texts: List[str]
                Input texts to be checked and optionally truncated.
            implicit_truncation_errors: bool
                Configuration indicates whether implicit truncation should be rejected.
        Returns:
            Tuple containing a dictionary of lists/arrays/tensors returned by the tokenizer, with
            proper truncation ('input_ids', 'attention_mask', etc.), and the input_token_count int.
        """

        max_tokens = self.max_seq_length

        # Do truncation if given a usable truncation value, else test for need to truncation
        if truncate_input_tokens < 0:
            okay_to_truncate = True
            max_length = max_tokens
        elif 0 < truncate_input_tokens <= max_tokens:
            okay_to_truncate = True
            # Add 2 for begin/end tokens, but don't go higher than model's max_tokens
            max_length = min(truncate_input_tokens + 2, max_tokens)

        else:
            okay_to_truncate = not implicit_truncation_errors
            max_length = max_tokens

        assert len(texts) > 0, "Cannot truncate nothing"
        assert isinstance(texts[0], str), "Only str can be truncated"

        to_tokenize = [texts]
        to_tokenize = [[str(s).strip() for s in col] for col in to_tokenize]
        tokenized = self.tokenizer(
            *to_tokenize,
            return_attention_mask=True,
            return_token_type_ids=False,
            return_overflowing_tokens=True,
            return_offsets_mapping=True,
            return_length=True,
            return_tensors="pt",
            truncation=True,
            padding=True,
            max_length=max_length,
        )

        # When truncation occurs multiple encodings are created for a single sample text
        was_truncated = len(tokenized.encodings) > len(to_tokenize[0])

        if not okay_to_truncate and was_truncated:
            # re-tokenize without truncation to eliminate the duplication of certain
            # special tokens (eg. [CLS] and [SEP]) with each overflow encoding.
            tokenized = self.tokenizer(
                *to_tokenize,
                return_attention_mask=True,
                return_token_type_ids=False,
                return_overflowing_tokens=True,
                return_offsets_mapping=True,
                return_length=True,
                return_tensors="pt",
                truncation=False,
                padding=True,
            )

            tokens = 0
            for encoding in tokenized.encodings:
                tokens = max(sum(encoding.attention_mask), tokens)
            error.log_raise(
                "<NLP08391926E>",
                ValueError(
                    f"Token sequence length is longer than the specified "
                    f"maximum sequence length for this model ({tokens} > {max_tokens})."
                ),
            )

        input_token_count = sum_token_count(tokenized, truncate_only=True)

        # Tokenize without overflow for batching and truncation to work together.
        tokenized = self.tokenizer(
            *to_tokenize,
            return_attention_mask=True,
            return_token_type_ids=False,
            return_overflowing_tokens=False,
            return_offsets_mapping=False,
            return_length=False,
            return_tensors="pt",
            truncation=True,
            padding=True,
            max_length=max_length,
        )

        return TruncatedTokensTuple(tokenized, input_token_count)

    def encode(
        self,
        sentences: Union[str, List[str]],
        batch_size: int = 32,
        show_progress_bar: bool = None,
        output_value: str = "sentence_embedding",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        device: str = None,
        normalize_embeddings: bool = False,
        truncate_input_tokens: int = 0,
        return_token_count: bool = False,
        implicit_truncation_errors: bool = True,
    ) -> Union[EmbeddingResultTuple, List[torch.Tensor], np.ndarray, torch.Tensor]:
        """
        Computes sentence embeddings

        :param sentences: the sentences to embed
        :param batch_size: the batch size used for the computation
        :param show_progress_bar: Ignored here. Added for compatibility with super API.
        :param output_value: Ignored here. Added for compatibility with super API.
        :param convert_to_numpy: If true, the output is a list of numpy vectors. Else, it is a list
                of pytorch tensors.
        :param convert_to_tensor: If true, you get one large tensor as return. Overwrites any
                setting from convert_to_numpy
        :param device: Which torch.device to use for the computation
        :param normalize_embeddings: Ignored here. Added for compatibility with super API.
        :param truncate_input_tokens: Truncation length for input tokens.
                Truncation length for input tokens.
                If less than zero, this truncation is left up to the tokenizer default (model max).
                If zero or greater than the model's maximum, then this is used as a test
                to see if truncation is needed. If truncation is needed, an exception is thrown,
                unless implicit_truncation_errors=False (see below).
                Otherwise, we take this usable truncation limit to truncate the input tokens.
        :param return_token_count: If true, a tuple is returned to add the input token count.
        :param implicit_truncation_errors: If true (default) implicit truncation throws an error.
                If false, the model default behavior or used.

        :return:
           If return_token_count is False, the embedding is returned as a numpy matrix.
           If return_token_count is True, a tuple is returned with both the embedding and
                the input token count.
        """

        # These args are for API compatability, but are currently ignored in our version of encode()
        _ = (
            show_progress_bar,
            output_value,
            normalize_embeddings,
        )

        self.eval()

        if convert_to_tensor:
            convert_to_numpy = False

        input_was_string = False
        list_of_sentences = sentences
        if isinstance(list_of_sentences, str) or not isinstance(
            sentences, Sized
        ):  # Cast an individual sentence to a list with length 1
            list_of_sentences = [sentences]
            input_was_string = True

        error.type_check_all("<NLP82314994E>", str, sentences=list_of_sentences)

        if device is None:
            device = self.device

        self.to(device)

        all_embeddings = []

        # Sort sentences according to length, from longest to shortest
        # OOM errors then occurs at start of encoding
        length_sorted_idx = np.argsort(
            [-self._text_length(sen) for sen in list_of_sentences]
        )
        sentences_sorted: list[str] = [
            list_of_sentences[idx] for idx in length_sorted_idx
        ]

        input_token_count = 0

        for start_index in range(0, len(list_of_sentences), batch_size):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features, token_count = self._truncate_input_tokens(
                truncate_input_tokens,
                sentences_batch,
                implicit_truncation_errors=implicit_truncation_errors,
            )
            input_token_count += token_count

            features = batch_to_device(features, device)

            if AUTOCAST:
                with torch.no_grad(), torch.cpu.amp.autocast():
                    out_features = self.forward(features)
                    embeddings = out_features["sentence_embedding"]
                    if convert_to_numpy:
                        embeddings = embeddings.detach().cpu()
                    all_embeddings.extend(embeddings)
            else:
                with torch.no_grad():
                    out_features = self.forward(features)
                    embeddings = out_features["sentence_embedding"]
                    if convert_to_numpy:
                        embeddings = embeddings.detach().cpu()
                    all_embeddings.extend(embeddings)

        # Restore original order
        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if convert_to_tensor:
            all_embeddings = torch.stack(all_embeddings)
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])

        if input_was_string:
            all_embeddings = all_embeddings[0]

        return (
            EmbeddingResultTuple(all_embeddings, input_token_count)
            if return_token_count
            else all_embeddings
        )



"""Tests for prompt tuning based inference via TGIS backend; note that these tests mock the
TGIS client and do NOT actually start/hit a TGIS server instance.
"""
# Standard
from typing import Iterable
from unittest import mock
import os
import tempfile

# Third Party
import pytest

# Local
from caikit_nlp.modules.text_generation import PeftPromptTuningTGIS
from tests.fixtures import (
    StubTGISClient,
    causal_lm_dummy_model,
    causal_lm_train_kwargs,
    saved_causal_lm_dummy_model,
    stub_tgis_backend,
    temp_config,
)

SAMPLE_TEXT = "Hello stub"


def test_load_and_run(causal_lm_dummy_model, stub_tgis_backend):
    """Ensure we can export an in memory model, load it, and (mock) run it with the right text & prefix ID."""
    # Patch our stub backend into caikit so that we don't actually try to start TGIS
    causal_lm_dummy_model.verbalizer = "hello distributed {{input}}"

    with mock.patch.object(StubTGISClient, "Generate") as mock_gen:
        mock_gen.side_effect = StubTGISClient.unary_generate

        # Save the local model & reload it a TGIS backend distributed module
        # Also, save the name of the dir + prompt ID, which is the path TGIS expects for the prefix ID
        with tempfile.TemporaryDirectory() as model_dir:
            causal_lm_dummy_model.save(model_dir)
            mock_tgis_model = PeftPromptTuningTGIS.load(model_dir, stub_tgis_backend)
            model_prompt_dir = os.path.split(model_dir)[-1]

        # Run an inference request, which is wrapped around our mocked Generate call
        result = mock_tgis_model.run(
            SAMPLE_TEXT, preserve_input_text=True, max_new_tokens=200, min_new_tokens=50
        )
        StubTGISClient.validate_unary_generate_response(result)

        stub_generation_request = mock_gen.call_args_list[0].args[0]

        # Validate that our verbalizer carried over correctly & was applied at inference time
        assert mock_tgis_model.verbalizer == causal_lm_dummy_model.verbalizer
        assert stub_generation_request.requests[
            0
        ].text == "hello distributed {}".format(SAMPLE_TEXT)

        # Ensure that our prefix ID matches the expected path based on our tmpdir and config
        assert model_prompt_dir == stub_generation_request.prefix_id


def test_load_and_tokenize(causal_lm_dummy_model, stub_tgis_backend):
    """Ensure we can export an in memory model, load it, and tokenize it"""
    # Patch our stub backend into caikit so that we don't actually try to start TGIS
    causal_lm_dummy_model.verbalizer = "hello distributed {{input}}"

    with mock.patch.object(StubTGISClient, "Tokenize") as mock_gen:
        mock_gen.side_effect = StubTGISClient.tokenize

        # Save the local model & reload it a TGIS backend distributed module
        with tempfile.TemporaryDirectory() as model_dir:
            causal_lm_dummy_model.save(model_dir)
            mock_tgis_model = PeftPromptTuningTGIS.load(model_dir, stub_tgis_backend)

    result = mock_tgis_model.run_tokenizer(SAMPLE_TEXT)
    StubTGISClient.validate_tokenize_response(result)

    # Validate that our verbalizer carried over correctly & was applied at inference time
    assert mock_tgis_model.verbalizer == causal_lm_dummy_model.verbalizer


def test_load_and_run_stream_out(causal_lm_dummy_model, stub_tgis_backend):
    """Ensure we can export an in memory model, load it, and (mock) run output streaming
    with the right text & prefix ID."""
    # Patch our stub backend into caikit so that we don't actually try to start TGIS
    causal_lm_dummy_model.verbalizer = "hello distributed {{input}}"

    with mock.patch.object(StubTGISClient, "GenerateStream") as mock_gen:
        mock_gen.side_effect = StubTGISClient.stream_generate

        # Save the local model & reload it a TGIS backend distributed module
        # Also, save the name of the dir + prompt ID, which is the path TGIS expects for the prefix ID
        with tempfile.TemporaryDirectory() as model_dir:
            causal_lm_dummy_model.save(model_dir)
            mock_tgis_model = PeftPromptTuningTGIS.load(model_dir, stub_tgis_backend)
            model_prompt_dir = os.path.split(model_dir)[-1]
            stub_tgis_backend.load_prompt_artifacts.assert_called_once()

        # Run an inference request, which is wrapped around our mocked GenerateStream call
        stream_result = mock_tgis_model.run_stream_out(
            SAMPLE_TEXT, preserve_input_text=True, max_new_tokens=200, min_new_tokens=50
        )
        StubTGISClient.validate_stream_generate_response(stream_result)

        stub_generation_request = mock_gen.call_args_list[0].args[0]

        # Validate that our verbalizer carried over correctly & was applied at inference time
        assert mock_tgis_model.verbalizer == causal_lm_dummy_model.verbalizer
        assert stub_generation_request.request.text == "hello distributed {}".format(
            SAMPLE_TEXT
        )

        # Ensure that our prefix ID matches the expected path based on our tmpdir and config
        assert model_prompt_dir == stub_generation_request.prefix_id


def test_purge_prompt_on_del(saved_causal_lm_dummy_model, stub_tgis_backend):
    """Test that the prompt artifacts get purged when a model is deleted"""

    # Load the model and make sure the prompt got copied over
    mock_tgis_model = PeftPromptTuningTGIS.load(
        saved_causal_lm_dummy_model, stub_tgis_backend
    )
    stub_tgis_backend.load_prompt_artifacts.assert_called_once()

    # Delete the model and make sure the prompt got "removed"
    with temp_config(unload_tgis_prompt_artifacts=True):
        mock_tgis_model.__del__()
        stub_tgis_backend.unload_prompt_artifacts.assert_called_once()
        prompt_id = os.path.basename(saved_causal_lm_dummy_model)
        stub_tgis_backend.unload_prompt_artifacts.assert_called_with(
            mock_tgis_model.base_model_name, prompt_id
        )


def test_purge_prompt_disabled_on_del(saved_causal_lm_dummy_model, stub_tgis_backend):
    """Test that the prompt artifacts are not purged if disabled"""

    # Load the model and make sure the prompt got copied over
    mock_tgis_model = PeftPromptTuningTGIS.load(
        saved_causal_lm_dummy_model, stub_tgis_backend
    )
    stub_tgis_backend.load_prompt_artifacts.assert_called_once()

    # Delete the model and make sure the prompt got "removed"
    with temp_config(unload_tgis_prompt_artifacts=False):
        mock_tgis_model.__del__()
        assert not stub_tgis_backend.unload_prompt_artifacts.called
